"""
File input and output.

This module contains classes implementing file input and output.

"""
from __future__ import absolute_import, print_function, division
from builtins import *

__author__ = 'Florian Krause <florian@expyriment.org>, \
Oliver Lindemann <oliver@expyriment.org>'
__version__ = ''
__revision__ = ''
__date__ = ''



import atexit
import os
try:
    import locale
except ImportError:
    locale = None  # Does not exist on Android
import codecs
import re
import sys
import uuid
import time
from time import strftime
from platform import uname
from itertools import combinations

from . import defaults
from .. import _internals
from .. import misc
from ..misc._timer import get_time
from ..misc import unicode2byte, byte2unicode, get_experiment_secure_hash, module_hashes_as_string
from ..misc import statistics
from ._input_output import Input, Output


class InputFile(Input):
    """A class implementing an input file."""

    def __init__(self, filename, encoding=None):
        """Create an input file.

        All lines in the specified text file will be read into a list of
        strings.

        Parameters
        ----------
        filename : str, optional
            name of the input file

        encoding : str, optional
            the encoding used to read the content of the file

        """

        self._filename = filename
        self._current_line = 1
        self._lines = []
        if not(os.path.isfile(self._filename)):
            raise IOError(u"The input file '{0}' does not exist.".format(
                self._filename))

        if encoding is None:
            with open(filename, 'r') as fl:
                first_line = fl.readline()
                encoding = re.findall("coding[:=]\s*([-\w.]+)", first_line)
                if encoding == []:
                    second_line = fl.readline()
                    encoding = re.findall("coding[:=]\s*([-\w.]+)",
                                          second_line)
                    if encoding == []:
                        encoding = [None]
        else:
            encoding = [encoding]
        with codecs.open(self._filename, 'rb', encoding[0],
                         errors='replace') as f:
            for line in f:
                self._lines.append(byte2unicode(line).rstrip('\r\n'))

    @property
    def filename(self):
        """Getter for filename."""

        return self._filename

    @property
    def current_line(self):
        """Getter for current_line."""

        return self._current_line

    @property
    def n_lines(self):
        """Getter for n_lines."""

        return len(self._lines)

    @property
    def lines(self):
        """Getter for lines."""

        return self._lines

    def get_line(self, line=None):
        """Get a specific line.

        If no line is given, the current line will be returned and the value
        of current_line will be increased by one. First line is line 1.

        Parameters
        ----------
        line : int, optional
            number of the line to get

        Returns
        -------
        line : str
            line as string or None if line does not exist

        """

        if line is not None and (line < 1 or line > len(self._lines)):
            return None

        if line is not None:
            return self._lines[line - 1]
        else:
            current_line = self._current_line
            if current_line != len(self._lines):
                self._current_line += 1
            return self._lines[current_line - 1]


class OutputFile(Output):
    """A class implementing an output file."""

    def __init__(self, suffix, directory, comment_char=None,
                 time_stamp=None):
        """Create an output file.

        Filename: {MAINFILE_NAME}_{SUBJECT_ID}_{TIME_STAMP}{suffix}

        Parameters
        ----------
        suffix : str
            file suffix/extension (str)
        directory : str
            create file in given directory
        comment_char : str, optional
            comment character
        time_stamp : bool, optional
            using time stamps, based on the experiment start time,
            not the current time

        """

        Output.__init__(self)
        self._suffix = suffix
        self._directory = directory
        if comment_char is not None:
            self._comment_char = comment_char
        else:
            self._comment_char = defaults.outputfile_comment_char
        if time_stamp is not None:
            self._time_stamp = time_stamp
        else:
            self._time_stamp = defaults.outputfile_time_stamp
        self._buffer = []
        if not os.path.isdir(directory):
            os.mkdir(directory)
        self._filename = self.standard_file_name
        self._fullpath = directory + "{0}{1}".format(os.path.sep, self._filename)

        atexit.register(self.save)

        # Create new file
        fl = open(self._fullpath, 'w+')
        fl.close()
        try:
            locale_enc = locale.getdefaultlocale()[1]
        except:
            locale_enc = "UTF-8"
        self.write_comment("Expyriment {0}, {1}-file, coding: {2}".format(
            _internals.get_version(), self._suffix,
            locale_enc))
        if _internals.active_exp.is_initialized:
            self.write_comment("date: {0}".format(time.strftime(
                               "%a %b %d %Y %H:%M:%S",
                               _internals.active_exp.clock.init_localtime)))

    @property
    def fullpath(self):
        """Getter for fullpath"""
        return self._fullpath

    @property
    def filename(self):
        """Getter for filename"""
        return self._filename

    @property
    def directory(self):
        """Getter for directory"""
        return self._directory

    @property
    def suffix(self):
        """Getter for directory"""
        return self._suffix

    @property
    def comment_char(self):
        """Getter for comment_char"""
        return self._comment_char

    @property
    def standard_file_name(self):
        """Getter for the standard expyriment outputfile name.

        Filename: {MAINFILE_NAME}_{SUBJECT_ID}_{TIME_STAMP}{suffix}

        """

        rtn = os.path.split(sys.argv[0])[1].replace(".py", "")
        if _internals.active_exp.is_started:
            rtn = rtn + '_' + repr(_internals.active_exp.subject).zfill(2)
        if self._time_stamp:
            rtn = rtn + '_' + strftime(
                "%Y%m%d%H%M", _internals.active_exp.clock.init_localtime)
        return rtn + self.suffix

    def save(self):
        """Save file to disk."""

        start = get_time()
        if self._buffer != []:
            buffer = [unicode2byte(x) for x in self._buffer]
            with open(self._fullpath, 'ab') as f:
                f.write(b"".join(buffer))
            self._buffer = []
        return int((get_time() - start) * 1000)

    def write(self, content):
        """Write to file.

        Parameters
        ----------
        content : str
            content to be written (anything, will be casted to str)

        """

        # NOTE: Do not print here recursion due to std_out logging

        if not _internals.is_base_string(content):
            content = str(content)
        self._buffer.append(content)


    def write_line(self, content):
        """Write a text line to files.

        Parameters
        ----------
        content : str
            content to be written (anything, will be casted to str)

        """
        self.write(content)
        self.write(defaults.outputfile_eol)

    def write_list(self, list_):
        """Write a list in a row. Data are separated by a delimiter.

        Parameters
        ----------
        list_ : list
            list to be written

        """

        for elem in list:
            self.write(elem)
            self.write(',')
        self.write(defaults.outputfile_eol)
        # self.write_line(repr(list_)[1:-1].replace(" ", ""))

    def write_comment(self, comment):
        """Write a comment line to files.

        (i.e., text is proceeded by comment char).

        Parameters
        ----------
        comment : str
            comment to be written (anything, will be casted to str)

        """

        self.write(self.comment_char)
        self.write_line(comment)

    def rename(self, new_filename):
        """Renames the output file."""
        self.save()
        new_fullpath = self.directory + "{0}{1}".format(os.path.sep, new_filename)
        os.rename(self._fullpath, new_fullpath)
        self._filename = new_filename
        self._fullpath = new_fullpath

    @staticmethod
    def get_next_subject_number():
        """Return the next subject number based on the existing data and event files."""

        subject_number = 1
        file_list = []
        if os.path.isdir(defaults.datafile_directory):
            file_list.extend(os.listdir(defaults.datafile_directory))
        if os.path.isdir(defaults.eventfile_directory):
            file_list.extend(os.listdir(defaults.eventfile_directory))

        if len(file_list)>0:
            mainfile_name = os.path.split(sys.argv[0])[1].replace(".py", "")
            for filename in file_list:
                if filename.startswith(mainfile_name) and \
                        (filename.endswith(DataFile._file_suffix) or
                         filename.endswith(EventFile._file_suffix)):
                    tmp = filename.replace(mainfile_name, "")\
                                  .replace(DataFile._file_suffix, "")\
                                  .replace(EventFile._file_suffix, "")
                    tmp = tmp.split('_')
                    try:
                        num = int(tmp[1])
                        if num >= subject_number and num<200000000000:
                            # large numbers are probably timestamps (normally the current event has no subject id
                            # yet and only a timestamp)
                            subject_number = num + 1
                    except:
                        pass
        return subject_number


class DataFile(OutputFile):
    """A class implementing a data file."""

    _file_suffix = ".xpd"

    def __init__(self, additional_suffix, directory=None, delimiter=None,
                 time_stamp=None):
        """Create a data file.

        Filename: {MAINFILE_NAME}_{SUBJECT_ID}_{TIME_STAMP}{ADD_SUFFIX}.xpd

        Parameters
        ----------
        additional_suffix : str
            additional suffix
        directory : str, optional
            directory of the file
        delimiter : str, optional
            symbol between variables
        time_stamp : bool, optional
            using time stamps, based on the experiment start time,
            not the current time

        """

        if _internals.active_exp.is_initialized:
            self._subject = _internals.active_exp.subject
        else:
            self._subject = None
        if directory is None:
            directory = defaults.datafile_directory
        if additional_suffix is None:
            additional_suffix = ''
        if len(additional_suffix) > 0:
            suffix = ".{0}{1}".format(additional_suffix, self._file_suffix)
        else:
            suffix = self._file_suffix
        OutputFile.__init__(self, suffix, directory, time_stamp=time_stamp)
        if delimiter is not None:
            self._delimiter = delimiter
        else:
            self._delimiter = defaults.datafile_delimiter

        self._subject_info = []
        self._experiment_info = []
        self._variable_names = []

        self.write_comment("--EXPERIMENT INFO")
        self.write_comment("e mainfile: {0}".format(os.path.split(
                                                    sys.argv[0])[1]))

        self.write_comment("e sha1: {0}".format(
                                    get_experiment_secure_hash()))
        self.write_comment("e modules: {0}".format(
                            module_hashes_as_string()))
        self.write_comment("--SUBJECT INFO")
        self.write_comment("s id: {0}".format(self._subject))
        self.write_comment("#")
        self._variable_names_changed = True
        self.save()

    @property
    def delimiter(self):
        """Getter for delimiter"""
        return self._delimiter

    def add(self, data):
        """Add data.

        Parameters
        ----------
        data : string or numeric or list
            data to be added

        """

        self.write(str(self._subject) + self.delimiter)
        if isinstance(data, (list, tuple)):
            line = ""
            for counter, elem in enumerate(data):
                if counter > 0:
                    line = line + self.delimiter
                if not _internals.is_base_string(elem):
                    elem = str(elem)
                if '"' in byte2unicode(elem):
                    elem = byte2unicode(elem).replace('"', '""')
                if ',' in byte2unicode(elem):
                    elem = '"{0}"'.format(byte2unicode(elem))
                line = line + elem
            self.write_line(line)
        else:
            if not _internals.is_base_string(data):
                data = str(data)
            self.write_line(data)

    def add_subject_info(self, text):
        """Adds a text the subject info header.

        Subject information can be extracted afterwards using
        misc.data_preprocessing.read_data_file. To defined between subject
        variables use a syntax like this: "gender = female" or
        "handedness : left"

        Parameters
        ----------
        text : str
            subject information to be added to the file header

        Notes
        -----
        The next data.save() might take longer!


        """

        self._subject_info.append(u"{0}s {1}{2}".format(
            self.comment_char, text, defaults.outputfile_eol))

    def add_experiment_info(self, text):
        """Adds a text the subject info header.

        Parameters
        ----------
        text : str
            experiment information to be added to the file header

        Notes
        -----
        The next data.save() might take longer!

        """

        for line in text.splitlines():
            self._experiment_info.append(u"{0}e {1}{2}".format(
                self.comment_char, line, defaults.outputfile_eol))

    @property
    def variable_names(self):
        """Getter for variable_names."""

        vn = self.delimiter.join(self._variable_names)
        return "subject_id,{0}".format(vn)

    def clear_variable_names(self):
        """Remove all variable names from data file.

        Notes
        -----
        The next data.save() might take longer!

        """

        self._variable_names = []
        self._variable_names_changed = True

    def add_variable_names(self, variable_names):
        """Add data variable names to the data file.

        Notes
        -----
        The next data.save() might take longer!

        Parameters
        ----------
        variables : str or list of str
            variable names

        """

        if variable_names is None:
            return
        try:
            variable_names = list(variable_names)
        except:
            variable_names = [variable_names]
        self._variable_names.extend(variable_names)
        self._variable_names_changed = True

    def save(self):
        """Save the new data to data-file.

        Returns
        -------
        time : int
            the time it took to execute this method

        """


        start = get_time()
        if len(self._subject_info) > 0 or len(self._experiment_info) > 0  \
                or self._variable_names_changed:
            # Re-write header and varnames
            tmpfile_name = "{0}{1}{2}".format(self.directory, os.path.sep, uuid.uuid4())
            os.rename(self._fullpath, tmpfile_name)
            fl = open(self._fullpath, 'wb+')
            tmpfl = open(tmpfile_name, 'r')
            section = None
            while True:
                line = tmpfl.readline()
                if not line:
                    break
                if line.startswith(self.comment_char + "e"):
                    section = "e"
                elif line.startswith(self.comment_char + "s"):
                    section = "s"
                else:
                    if section == "e":  # Previous line was last #e
                        if len(self._experiment_info) > 0:
                            fl.write(unicode2byte("".join(self._experiment_info)))
                            self._experiment_info = []
                        section = None
                    elif section == "s":  # Previous line was last #s
                        if len(self._subject_info) > 0:
                            fl.write(unicode2byte("".join(self._subject_info)))
                            self._subject_info = []
                        section = None

                        # Re-write variable names after #s-section
                        fl.write(unicode2byte(
                            self.variable_names + defaults.outputfile_eol))
                        self._variable_names_changed = False
                        line = ''  # Skip old varnames
                fl.write(unicode2byte(line))
            tmpfl.close()
            fl.close()

            os.remove(tmpfile_name)
            self._subject_info = []
            self._experiment_info = []

        if self._buffer != []:
            OutputFile.save(self)
            if self._logging:
                _internals.active_exp._event_file_log("Data,saved")

        return int((get_time() - start) * 1000)


class EventFile(OutputFile):
    """A class implementing an event file."""

    _file_suffix = ".xpe"

    def __init__(self, additional_suffix, directory=None, delimiter=None,
                 clock=None, time_stamp=None):
        """Create an event file.

        Filename: {MAINFILE_NAME}_{SUBJECT_ID}_{TIME_STAMP}{ADD_SUFFIX}.xpd

        Parameters
        ----------
        additional_suffix : str
            additional suffix
        directory : str, optional
            directory of the file
        delimiter : str, optional
            symbol between timestamp and event
        clock : expyriment.Clock, optional
            an experimental clock
        time_stamp : bool, optional
            using time stamps, based on the experiment start time,
            not the current time

        """

        if directory is None:
            directory = defaults.eventfile_directory
        if additional_suffix is None:
            additional_suffix = ''
        if len(additional_suffix) > 0:
            suffix = ".{0}{1}".format(additional_suffix, self._file_suffix)
        else:
            suffix = self._file_suffix
        OutputFile.__init__(self, suffix, directory, time_stamp=time_stamp)
        if delimiter is not None:
            self._delimiter = delimiter
        else:
            self._delimiter = defaults.eventfile_delimiter
        if clock is not None:
            self._clock = clock
        else:
            if not _internals.active_exp.is_initialized:
                raise RuntimeError(
                    "Cannot find a clock. Initialize Expyriment!")
            self._clock = _internals.active_exp.clock

        try:
            display = repr(_internals.active_exp.screen.window_size)
            window_mode = repr(_internals.active_exp.screen.window_mode)
            open_gl = repr(_internals.active_exp.screen.open_gl)
        except:
            display = "unknown"
            window_mode = "unknown"
            open_gl = "unknown"

        self.write_comment("sha1: {0}".format(
                                    get_experiment_secure_hash()))
        self.write_comment("modules: {0}".format(
                            module_hashes_as_string()))
        self.write_comment("display: size={0}, window_mode={1}, open_gl={2}".format(
            display, window_mode, open_gl))
        self.write_comment("os: {0}".format(uname()))

        self.write_line("Time,Type,Event,Value,Detail,Detail2")
        self.save()

        self._inter_event_intervall_log = _InterEventIntervallLog()
        atexit.register(self._write_inter_event_intervall_summary)

    @property
    def clock(self):
        """Getter for clock"""
        return self._clock

    @property
    def delimiter(self):
        """Getter for delimiter"""
        return self._delimiter

    def log(self, event, log_event_tag=None):
        """Log an event.

        Parameters
        ----------
        event : anything
            the event to be logged (anything, will be casted to str)
        log_event_tag : numeral or string, optional
            if log_event_tag is defined, event file logs the inter-event-intervalls
            and adds a summary of the intervalls at the end of the file

        Returns
        -------
        log_time : int
            the time of logging

        """

        log_time = self._clock.time
        if not _internals.is_base_string(event):
            event = str(event)
        self.write_line(repr(log_time) + self.delimiter + event)
        if log_event_tag is not None:
            self._inter_event_intervall_log.add_event(log_event_tag, log_time)
        return log_time

    def warn(self, message):
        """Log a warning message.

        Parameters
        ----------
        message : str
            warning message to log

        """

        line = "WARNING: " + message
        self.write_line(line)

    def _write_inter_event_intervall_summary(self):
        """appending the inter event intervall summary to event file, if log_event_tag have been set while presentation
        this function will be called at exit"""

        for l in self._inter_event_intervall_log.summary():
            self.write_comment(l)



class _InterEventIntervallLog():
    """This class is used to log the intervalls of tagged events to get a
    summary of the timing at the end of the event file
    """

    def __init__(self):
        self.clear()


    def clear(self):
        """clear logging queue"""
        self.log_dict = {}


    def add_event(self, event_tag, time):
        """add event for logging"""

        try:
            self.log_dict[event_tag].append(time)
        except:
            self.log_dict[event_tag] = [time]


    def _get_iei_intervalls(self, from_tag, to_tag):
        """helper function: get the intervalls between two events"""

        rtn = []
        try:
            time_from = sorted(self.log_dict[from_tag])
            time_to = sorted(self.log_dict[to_tag])
        except:
            return rtn

        for f in time_from:
            # find first larger to time and add different to list
            for t in time_to:
                if t>=f:
                    rtn.append(t-f)
                    break
        return rtn


    def summary(self):
        """The summary as string array

        Returns
        -------
        txt : string array

        """

        rtn = []
        for a, b in combinations(self.log_dict.keys(), 2):
            for reverse in [False, True]:
                if reverse:
                    tmp = b
                    b = a
                    a = tmp
                iei = self._get_iei_intervalls(a, b)
                txt = "{0} --> {1}: n={2}".format(a,b, len(iei))
                if len(iei)>0:
                    txt += ", mean={0}, median={1}, std={2}".format(
                                misc.round(statistics.mean(iei),2),
                                misc.round(statistics.median(iei),2),
                                misc.round(statistics.std(iei),2),
                                )
                rtn.append(txt)
        return rtn



"""
Mouse input.

This module contains a class implementing pygame mouse input.

"""
from __future__ import absolute_import, print_function, division
from builtins import *


__author__ = 'Florian Krause <florian@expyriment.org>, \
Oliver Lindemann <oliver@expyriment.org>'
__version__ = ''
__revision__ = ''
__date__ = ''


from types import FunctionType

import pygame

from . import defaults
from ..misc._timer import get_time
from ..misc import is_android_running
from ._input_output  import Input
from .. import _internals, misc

class Mouse(Input):
    """A class implementing a mouse input.

    Calling ``expyriment.control.initialize(exp)`` will automatically create a
    mouse instance and will reference it in exp.mouse for easy access.

    """

    #static class properties for quit_events
    _quit_corner_location = None
    _corner_rect_size = (30, 30)
    _quit_action_events = []


    def __init__(self, show_cursor=True, track_button_events=None,
                 track_motion_events=None):
        """Initialize a mouse input.

        Parameters
        ----------
        show_cursor : bool, optional
            shows mouse cursor (default = True)
        track_button_events : bool, optional
            track button events via Pygame queue (default = True)
        track_motion_events : bool, optional
            track motion events via Pygame queue (default = False)

        Notes
        -----
        (a) It is strongly suggest to avoid tracking of motions events,
        (track_motion_events=True), because it quickly causes an overflow in
        the Pygame event queue and you might consequently loose important
        events.

        (b) Turning the mouse wheel causes button_down_events. Thus,
        turning the mouse wheel rather extensively causes an overflow of
        the Pygame event queue. You might consider turn off the default
        tracking of mouse button event by calling
        `experiment.mouse.track_button_events = False`.

        (c) See ``process_quit_event`` for the forced quitting of experiments
        via mouse events.

        """

        Input.__init__(self)
        if is_android_running():
            Mouse._quit_corner_location = 1
        if show_cursor is None:
            show_cursor = defaults.mouse_show_cursor
        if track_button_events is None:
            track_button_events = defaults.mouse_track_button_events
        if track_motion_events is None:
            track_motion_events = defaults.mouse_track_motion_events
        if show_cursor:
            self.show_cursor(track_button_events, track_motion_events)
        else:
            self.track_button_events = track_button_events
            self.track_motion_events = track_motion_events

    @staticmethod
    def set_quit_corner_location(corner, corner_rect_size=(None, None)):
        """Activate the possibility to quit experiment using a mouse event.

        Defines the corner on which the user has to click to elicit a
        quit dialog.

        If quit corner location has been defined, clicking quickly three
        times (i.e., within 1 second) in the specified corner forces
        experiment to quit.

        To switch off the detection of mouse quit events, please call
        ``Mouse.set_quit_corner_location(corner=None)``.

        Changing the corner and rect_size always affects all mouse
        instances.

        Parameters
        ----------
        corner: int or None
            location code (0,1, 2 or 3) of the quit corner; the default value
            under Android is 1, otherwise None; see also the notes
            below.

       corner_rect_size = tuple (int, int), optional
            size of the field (rect) that detects the quit action in one
            corner of the screen; default = (30, 30)

        Notes
        -----
        Mouse quit events are especially useful for experiments on devices
        without hardware keyboard, such as tablet PCs or smartphones.

        Corner location codes::

                0 = upper left corner,  1 = upper right corner   (0) (1)
                2 = lower right corner, 3 = lower left corner    (3) (2)
                otherwise the detection of mouse quit events is deactivated

        The detection of mouse quit events is activated by default under
        Android.

        """

        if corner is not None:
            if not isinstance(corner, int) or corner<0 or corner >3:
                corner = None
                print("Warning: {} is an unkown corner location. Mouse quit "
                                     "event is deactivated.".format(corner))
        Mouse._quit_corner_location = corner

        try:
            Mouse._corner_rect_size = (int(corner_rect_size[0]),
                                       int(corner_rect_size[1]))
        except:
            pass

    @staticmethod
    def process_quit_event(click_position=None):
        """Check if mouse exit action has been performed.

        If quit corner location is has been defined via
        ``Mouse.set_quit_corner_location()``  (i.e. 0, 1, 2 or 3), clicking
        quickly three times (i.e., within 1 second) in one of the corners of
        the screen forces the experiment to quit.

        The function is called automatically by all mouse get event and wait
        methods (similar to ``Keyboard.process_control_keys``).  If no mouse
        functions are called by your program, this function can be polled to
        ensure quitting experiment by mouse.

        Parameters
        ----------
        click_position : tuple of int (x,y), optional
            clicked location to be processed. If not defined, the Pygame event
            queue will be checked for mouse down events and the current
            position is taken

        Returns
        -------
        out : bool, optional
            True if exit action has been performed
            False otherwise

        See Also
        --------
        set_quit_corner_location

        """

        if Mouse._quit_corner_location not in (0, 1, 2, 3):
            return False

        if click_position is None:
            # check Pygame queu
            pos = None
            # pygame.event.pump() # not sure if it is required!
            for event in pygame.event.get(pygame.MOUSEBUTTONDOWN):
                if event.button > 0:
                    screen_size = _internals.active_exp.screen.surface.get_size()
                    pos = pygame.mouse.get_pos()
                    pos = (pos[0] - screen_size[0] // 2,
                          -pos[1] + screen_size[1] // 2)
                    break
            if pos is None:
                return False
            else:
                return Mouse.process_quit_event(click_position=pos)

        # determine threshold x & y
        if Mouse._quit_corner_location == 0 or Mouse._quit_corner_location == 3: # left
            threshold_x = -_internals.active_exp.screen.center_x + \
                    Mouse._corner_rect_size[0]
        else:# right
            threshold_x = _internals.active_exp.screen.center_x - \
                    Mouse._corner_rect_size[0]
        if Mouse._quit_corner_location == 0 or Mouse._quit_corner_location == 1: # upper
            threshold_y = _internals.active_exp.screen.center_y - \
                    Mouse._corner_rect_size[1]
        else:# lower
            threshold_y = -_internals.active_exp.screen.center_y + \
                    Mouse._corner_rect_size[1]
        # check
        if (Mouse._quit_corner_location == 0 and \
            click_position[0] < threshold_x and \
            click_position[1] > threshold_y) \
           or (Mouse._quit_corner_location == 1 and \
               click_position[0] > threshold_x and \
               click_position[1] > threshold_y) \
           or (Mouse._quit_corner_location == 2 and \
               click_position[0] > threshold_x and \
               click_position[1] < threshold_y) \
           or (Mouse._quit_corner_location == 3 and \
               click_position[0] < threshold_x and \
               click_position[1] < threshold_y):

            Mouse._quit_action_events.append(get_time())
            if len(Mouse._quit_action_events)>=3:
                diff = get_time()-Mouse._quit_action_events.pop(0)
                if (diff < 1):
                    # simulate quit key
                    simulated_key = pygame.event.Event(
                        pygame.KEYDOWN,
                        {'key': _internals.active_exp.keyboard.get_quit_key()})
                    return _internals.active_exp.keyboard.process_control_keys(
                        key_event=simulated_key)
        return False


    @property
    def track_button_events(self):
        """Getter for track_button_events."""

        return self._track_button_events

    @track_button_events.setter
    def track_button_events(self, value):
        """Setter for track_button_events.

        Switch on/off the processing of button and wheel events.

        """

        self._track_button_events = value
        if value:
            pygame.event.set_allowed(pygame.MOUSEBUTTONDOWN)
            pygame.event.set_allowed(pygame.MOUSEBUTTONUP)
        else:
            pygame.event.set_blocked(pygame.MOUSEBUTTONDOWN)
            pygame.event.set_blocked(pygame.MOUSEBUTTONUP)

    @property
    def track_motion_events(self):
        """Getter for track_motion_events.

        Switch on/off the buffering of motion events in the Pygame event queue.

        Notes
        -----
        It is strongly suggest to avoid tracking of motions events,
        (track_motion_events=True), because it quickly causes an overflow in
        the Pygame event queue and you might consequently loose important
        events.

        """

        return self._track_motion_events

    @track_motion_events.setter
    def track_motion_events(self, value):
        """Setter for track_motion_events.

        Switch on/off the processing of motion events.

        """

        self._track_motion_events = value
        if value:
            pygame.event.set_allowed(pygame.MOUSEMOTION)
        else:
            pygame.event.set_blocked(pygame.MOUSEMOTION)

    @property
    def pressed_buttons(self):
        """Getter for pressed_buttons."""

        pygame.event.pump()
        return pygame.mouse.get_pressed()

    @property
    def is_cursor_visible(self):
        """Getter for is_cursor_visible"""

        visible = pygame.mouse.set_visible(False)
        pygame.mouse.set_visible(visible)
        return visible

    def get_last_button_down_event(self, process_quit_event=True):
        """Get the last button down event.
        All earlier button down events will be removed from the queue.

        Parameters
        ----------
        process_quit_event : boolean, optional
            if False, the current location will not be processed for mouse
            quitting events in the case that a button down event has been
            found (default = True).

        Returns
        -------
        btn_id : int
            button number (0,1,2) or 3 for wheel up or 4 for wheel down,
            if quit screen mouse action has been performed, the method
            returns -1

        """

        rtn = None
        for event in pygame.event.get(pygame.MOUSEBUTTONDOWN):
            if event.button > 0:
                rtn = event.button - 1
        if rtn==0:
            if process_quit_event and Mouse.process_quit_event(self.position):
                return -1
        return rtn

    def get_last_button_up_event(self):
        """Get the last button up event.
        All earlier button up events will be removed from the queue.

        Returns
        -------
        btn_id : int
            button number (0,1,2)
            if quit screen mouse action has been performed, the method
            returns -1

        """

        rtn = None
        for event in pygame.event.get(pygame.MOUSEBUTTONUP):
            if event.button > 0:
                rtn = event.button - 1
        return rtn

    def check_button_pressed(self, button_number):
        """Return False or button id if a specific button is currently pressed.

        Parameters
        ----------
        button_number : int
            the button number (0,1,2) to be checked

        Returns
        -------
        is_pressed: boolean

        """

        btns = self.pressed_buttons
        if len(btns) >= 1 and button_number >= 0:
            return btns[button_number]
        else:
            return False

    def check_wheel(self):
        """Check the mouse wheel.

        Returns
        -------
        direction : str
            "up" or "down" if mouse wheel has been recently rotated
            upwards or downwards otherwise it returns None.

        """

        evt = self.get_last_button_down_event()
        if evt == 3:
            return "up"
        elif evt == 4:
            return "down"
        else:
            return None

    @property
    def position(self):
        """Getter for position."""

        pygame.event.pump()
        screen_size = _internals.active_exp.screen.surface.get_size()
        pos = pygame.mouse.get_pos()
        return (pos[0] - screen_size[0] // 2, -pos[1] + screen_size[1] // 2)

    @position.setter
    def position(self, position):
        """Setter for position."""

        screen_size = _internals.active_exp.screen.surface.get_size()
        pos = (position[0] + screen_size[0] // 2,
               - position[1] + screen_size[1] // 2)
        pygame.mouse.set_pos(pos)

    def set_cursor(self, size, hotspot, xormasks, andmasks):
        """Set the cursor.

        Parameters
        ----------
        size : (int, int)
            size of the cursor
        hotspot : (int, int)
            position of the hotspot (0,0 is top left)
        xormask : list
            sequence of bytes with cursor xor data masks
        andmask : list
            sequence of bytes with cursor bitmask data

        """

        return pygame.mouse.set_cursor(size, hotspot, xormasks, andmasks)

    def get_cursor(self):
        """Get the cursor."""

        return pygame.mouse.get_cursor()

    def clear(self):
        """Clear the event cue from mouse events."""

        pygame.event.clear(pygame.MOUSEBUTTONDOWN)
        pygame.event.clear(pygame.MOUSEBUTTONUP)
        pygame.event.clear(pygame.MOUSEMOTION)
        if self._logging:
            _internals.active_exp._event_file_log("Mouse,cleared", 2)


    def wait_event(self, wait_button=True, wait_motion=True, buttons=None,
                   duration=None, wait_for_buttonup=False,
                   callback_function=None, process_control_events=True):
        """Wait for a mouse event (i.e., motion, button press or wheel event).

        Button id coding:

        - None    for no mouse button event or
        - 0,1,2   for left. middle and right button or
        - 3       for wheel up or
        - 4       for wheel down (wheel works only for keydown events).

        Parameters
        ----------
        wait_button : bool, optional
            set 'False' to ignore for a button presses (default=True)
        wait_motion : bool, optional
            set 'False' to ignore for a mouse motions (default=True)
        buttons : int or list, optional
            a specific button or list of buttons to wait for
        duration : int, optional
            the maximal time to wait in ms
        wait_for_buttonup : bool, optional
            if True it waits for button-up default=False)
        callback_function : function, optional
            function to repeatedly execute during waiting loop
        process_control_events : bool, optional
            process ``io.keyboard.process_control_keys()`` and
            ``io.mouse.process_quit_event()`` (default = True)

        Returns
        -------
        event_id : int
            id of the event that quited waiting
        move : bool
            True if a motion occurred
        pos : (int, int)
            mouse position (tuple)
        rt : int
            reaction time

        Notes
        ------
        This will also by default process control events (quit and pause).
        Thus, keyboard events will be cleared from the cue and cannot be
        received by a ``Keyboard().check()`` anymore!

        See Also
        --------
        expyriment.design.Experiment.register_wait_callback_function

        """

        if _internals.skip_wait_methods:
            return None, None, None, None
        start = get_time()
        self.clear()
        old_pos = pygame.mouse.get_pos()
        btn_id = None
        rt = None
        motion_occured = False
        if buttons is None:
            buttons = [0, 1, 2, 3, 4]
        else:
            try:
                buttons = list(buttons)
            except:
                buttons = [buttons]
        while True:
            if isinstance(callback_function, FunctionType):
                rtn_callback = callback_function()
                if isinstance(rtn_callback, _internals.CallbackQuitEvent):
                    btn_id = rtn_callback
                    rt = int((get_time() - start) * 1000)
                    break
            if _internals.active_exp.is_initialized:
                rtn_callback = _internals.active_exp._execute_wait_callback()
                if isinstance(rtn_callback, _internals.CallbackQuitEvent):
                    btn_id = rtn_callback
                    rt = int((get_time() - start) * 1000)
                    break
                if process_control_events:
                    if _internals.active_exp.keyboard.process_control_keys():
                        break
            if wait_motion:
                motion_occured = old_pos != pygame.mouse.get_pos()
            if wait_button:
                if wait_for_buttonup:
                    btn_id = self.get_last_button_up_event()
                else:
                    btn_id = self.get_last_button_down_event(
                        process_quit_event=process_control_events)
            if btn_id ==-1:
                btn_id = None
                break
            elif btn_id in buttons or motion_occured:
                rt = int((get_time() - start) * 1000)
                break
            elif (duration is not None and \
                    int((get_time() - start) * 1000) >= duration):
                break

        position_in_expy_coordinates = self.position

        if self._logging:
            _internals.active_exp._event_file_log(
            "Mouse,received,{0}-{1},wait_event".format(btn_id, motion_occured))
        return btn_id, motion_occured, position_in_expy_coordinates, rt


    def wait_press(self, buttons=None, duration=None, wait_for_buttonup=False,
                   callback_function=None, process_control_events=True):
        """Wait for a mouse button press or mouse wheel event.

        Parameters
        ----------
        buttons : int or list, optional
            a specific button or list of buttons to wait for
        duration : int, optional
            maximal time to wait in ms
        wait_for_buttonup : bool, optional
            if True it waits for button-up
        callback_function : function, optional
            function to repeatedly execute during waiting loop
        process_control_events : bool, optional
            process ``io.keyboard.process_control_keys()`` and
            ``io.mouse.process_quit_event()`` (default = false)

        Returns
        -------
        event_id : int
            id of the event that quited waiting
        pos : (int, int)
            mouse position (tuple)
        rt : int
            reaction time

        Notes
        ------
        button id coding

        - None    for no mouse button event or
        - 0,1,2   for left. middle and right button or
        - 3       for wheel up or
        - 4       for wheel down (wheel works only for keydown events).

        """

        rtn = self.wait_event(wait_button=True, wait_motion=False,
                              buttons=buttons, duration=duration,
                              wait_for_buttonup=wait_for_buttonup,
                              callback_function=callback_function,
                              process_control_events=process_control_events)
        return rtn[0], rtn[2], rtn[3]

    def wait_motion(self, duration=None, callback_function=None,
                    process_control_events=True):
        """Wait for a mouse motion.

        Parameters
        ----------
        duration : int, optional
            maximal time to wait in ms
        callback_function : function, optional
            function to repeatedly execute during waiting loop
        process_control_events : bool, optional
            process ``io.keyboard.process_control_keys()`` and
            ``io.mouse.process_quit_event()`` (default = false)

        Returns
        -------
        pos : (int, int)
            mouse position (tuple)
        rt : int
            reaction time

        """

        rtn = self.wait_event(wait_button=False, wait_motion=True, buttons=[],
                              duration=duration, wait_for_buttonup=False,
                              callback_function=callback_function,
                              process_control_events=process_control_events)

        if isinstance(rtn[0], _internals.CallbackQuitEvent):
            return rtn[0], rtn[3]
        else:
            return rtn[2], rtn[3]


    def show_cursor(self, track_button_events=True, track_motion_events=False):
        """Show the cursor.

        Parameters
        ----------
        track_button_events : bool, optional
            tracking button events (default = True)
        track_motion_events : bool, optional
            tracking motion events (default = False)

        """

        pygame.mouse.set_visible(True)
        self.track_button_events = track_button_events
        self.track_motion_events = track_motion_events

    def hide_cursor(self, track_button_events=False, track_motion_events=False):
        """Hide the cursor.

        Parameters
        ----------
        track_button_events : bool, optional
            tracking button events (default = False)
        track_motion_events : bool, optional
            tracking motion events (default = False)

        """

        pygame.mouse.set_visible(False)
        self.track_button_events = track_button_events
        self.track_motion_events = track_motion_events


    @staticmethod
    def _self_test(exp):
        """Test the mouse.

        Returns
        -------
        polling_time : int
            polling time
        buttons_work : int
            1 -- if mouse test was ended with mouse button,
            0 -- if testing has been quit with q

        """

        from .. import stimuli
        # measure mouse polling time
        info = """This will test how timing accurate your mouse is.

[Press RETURN to continue]"""

        stimuli.TextScreen("Mouse test (1)", info).present()
        exp.keyboard.wait(misc.constants.K_RETURN)
        mouse = Mouse()
        go = stimuli.TextLine("Keep on moving...")
        go.preload()
        stimuli.TextLine("Please move the mouse").present()
        mouse.wait_motion()
        go.present()
        exp.clock.reset_stopwatch()
        motion = []
        while exp.clock.stopwatch_time < 200:
            _pos, rt = mouse.wait_motion()
            motion.append(rt)
        stimuli.TextLine("Thanks").present()
        polling_time = misc.statistics.mode(motion)

        info = """Your mouse polling time is {0} ms.

[Press RETURN to continue] """.format(polling_time)
        text = stimuli.TextScreen("Results", info)
        text.present()
        exp.keyboard.wait([misc.constants.K_RETURN])

        info = """This will test if you mouse buttons work.
Please press all buttons one after the other to see if the corresponding buttons on the screen light up.
When done, click inside one of the buttons on the screen to end the test.
If your mouse buttons do not work, you can quit by pressing q.

[Press RETURN to continue]"""

        stimuli.TextScreen("Mouse test (2)", info).present()
        exp.keyboard.wait(misc.constants.K_RETURN)

        # test mouse clicking
        rects = [stimuli.Rectangle(size=[30, 30], position=[-50, 0]),
                 stimuli.Rectangle(size=[30, 30], position=[0, 0]),
                 stimuli.Rectangle(size=[30, 30], position=[50, 0])]
        canvas = stimuli.Canvas(size=[350, 500])
        btn = None
        go_on = True
        while go_on:
            canvas.clear_surface()
            for cnt, r in enumerate(rects):
                r.unload()
                if cnt == btn:
                    r.colour = misc.constants.C_YELLOW
                else:
                    r.colour = misc.constants.C_RED
                r.plot(canvas)

            if btn == 3:
                text = "Mouse wheel UP"
            elif btn == 4:
                text = "Mouse wheel DOWN"
            else:
                text = ""
            stimuli.TextLine(text, position=[0, 50]).plot(canvas)
            canvas.present()
            btn = None
            while btn is None:
                btn = mouse.get_last_button_down_event()
                if btn is not None:
                    position = mouse.position
                    for r in rects:
                        if r.overlapping_with_position(position):
                            buttons_work = 1
                            mouse.hide_cursor()
                            go_on = False
                            break
                elif exp.keyboard.check(keys=misc.constants.K_q):
                    buttons_work = 0
                    mouse.hide_cursor()
                    go_on = False
                    break

        result = {}
        result["testsuite_mouse_polling_time"] = str(polling_time) + " ms"
        result["testsuite_mouse_buttons_work"] = buttons_work
        return result




#!/usr/bin/env python

"""
This module contains the base classes for stimuli.

"""
from __future__ import absolute_import, print_function, division
from builtins import *

__author__ = 'Florian Krause <florian@expyriment.org>, \
Oliver Lindemann <oliver@expyriment.org>'
__version__ = ''
__revision__ = ''
__date__ = ''


from copy import deepcopy

from .._internals import Expyriment_object
from .. import _internals
from ..misc import byte2unicode

class Stimulus(Expyriment_object):
    """A class implementing a very general experimental stimulus.

    All other stimulus classes are based on this one.
    If a new stimulus is to be created, (at least) this class should be
    subclassed.

    """

    _id_counter = 0

    def __init__(self, log_comment=None):
        """Create a stimulus.

        Parameters
        ----------
        log_comment : str, optional
            comment for the event log file

        """

        Expyriment_object.__init__(self)
        self._id = Stimulus._id_counter
        Stimulus._id_counter += 1

        log_txt = u"Stimulus,created,{0},{1}".format(self.id,
                                                    self.__class__.__name__)

        if log_comment is not None:
            log_txt = u"{0},{1}".format(log_txt, byte2unicode(log_comment))
        if self._logging:
            _internals.active_exp._event_file_log(log_txt, 2)


    @property
    def id(self):
        """Getter for id."""

        return self._id

    def copy(self):
        """Return a deep copy of the stimulus."""

        copy = deepcopy(self)
        copy._id = Stimulus._id_counter
        Stimulus._id_counter += 1

        if self._logging:
            _internals.active_exp._event_file_log(
                    "Stimulus,created,{0},{1},copied from {2}".format(
                    copy.id, copy.__class__.__name__, self.id))
        return copy



"""The expyriment randomize module.

This module contains various functions for randomizing data

"""

__author__ = 'Florian Krause <florian@expyriment.org>,\
              Oliver Lindemann <oliver@expyriment.org>'
__version__ = ''
__revision__ = ''
__date__ = ''


from ._randomize import rand_int_sequence, rand_int, rand_element, rand_norm
from ._randomize import coin_flip, shuffle_list, make_multiplied_shuffled_list



#!/usr/bin/env python

"""
A Canvas stimulus.

This module contains a class implementing a canvas stimulus.

"""
from __future__ import absolute_import, print_function, division
from builtins import *

__author__ = 'Florian Krause <florian@expyriment.org>, \
Oliver Lindemann <oliver@expyriment.org>'
__version__ = ''
__revision__ = ''
__date__ = ''


import pygame

from . import defaults
from ._visual import Visual


class Canvas(Visual):
    """A class implementing a canvas stimulus."""

    def __init__(self, size, position=None, colour=None):
        """Create a canvas.

        Parameters
        ----------
        size : (int, int)
            size of the canvas (int,int)
        position : (int, int), optional
            position of the stimulus
        colour : (int,int,int), optional
            colour of the canvas stimulus

        """

        if position is None:
            position = defaults.canvas_position
        Visual.__init__(self, position)
        self._size = size
        if colour is not None:
            self._colour = colour
        else:
            self._colour = defaults.canvas_colour

    _getter_exception_message = "Cannot set {0} if surface exists!"

    @property
    def size(self):
        """Getter for size."""
        return self._size

    @size.setter
    def size(self, value):
        """Setter for size."""

        if self.has_surface:
            raise AttributeError(Canvas._getter_exception_message.format(
                "size"))
        else:
            self._size = value

    @property
    def colour(self):
        """Getter for colour."""
        return self._colour

    @colour.setter
    def colour(self, value):
        """Setter for colour."""

        if self.has_surface:
            raise AttributeError(Canvas._getter_exception_message.format(
                "colour"))
        else:
            self._colour = value

    def _create_surface(self):
        """Create the surface of the stimulus."""

        surface = pygame.surface.Surface(self._size,
                                         pygame.SRCALPHA).convert_alpha()
        if self._colour is not None:
            surface.fill(self._colour)
        return surface


    @staticmethod
    def _demo(exp=None):
        if exp is None:
            from .. import control
            control.set_develop_mode(True)
            control.defaults.event_logging = 0
            exp_ = control.initialize()
        cnvs = Canvas((200, 200), colour=(255, 255, 255))
        cnvs.present()
        if exp is None:
            exp_.clock.wait(1000)



"""Expyriment internal function and variables

This module also contains the currently active experiment:
            active_exp
"""

from builtins import object

__author__ = 'Florian Krause <florian@expyriment.org> \
Oliver Lindemann <oliver@expyriment.org>'
__version__ = ''
__revision__ = ''
__date__ = ''


import sys
import os
import pygame

try:
    import android
except ImportError:
    android = None


def get_version():
    """
    Return version information about Expyriment and Python.

    Returns
    -------
    version_info : str

    Notes
    -----
    For more detailed information see expyriment.misc.get_system_info().

    """

    pv = "{0}.{1}.{2}".format(sys.version_info[0],
                              sys.version_info[1],
                              sys.version_info[2])
            #no use of .major, .minor to ensure MacOS compatibility
    return "{0} (Python {1})".format(__version__, pv)


# GLOBALLY NEEDED STUFF

PYTHON3 = (sys.version_info[0] == 3)

active_exp = None  # expyriment.design.__init__ sets active_exp to
                   # design.Experiment("None")
                   # Provides the access to the currently active experiment
                   # import ._internals to read and write _active.exp

skip_wait_methods = False  # global toggle, can be changed by set_develop_mode

def is_base_string(s):
    if PYTHON3:
        return isinstance(s, (str, bytes))
    else:
        return isinstance(s, (unicode, str))

def is_unicode_string(s):
    if PYTHON3:
        return isinstance(s, str)
    else:
        return isinstance(s, unicode)

def is_byte_string(s):
    if PYTHON3:
        return isinstance(s, bytes)
    else:
        return isinstance(s, str)

def pump_pygame_events():
    pygame.event.pump()

class Expyriment_object(object):
    """A class implementing a general Expyriment object.

       Parent of all stimuli and IO objects.

    """

    def __init__(self):
        """Create an Expyriment object."""
        self._logging = True

    def set_logging(self, onoff):
        """Set logging of this object on or off

        Parameters
        ----------
        onoff : bool
            set logging on (True) or off (False)

        See Also
        --------
        expyriment.design.Experiment.set_log_level

        """

        self._logging = onoff

    @property
    def logging(self):
        """Getter for logging."""

        return self._logging

class CallbackQuitEvent(object):
    """A CallbackQuitEvent.

    If a callback function returns a CallbackQuitEvent object the currently processed
    the wait or event loop function will be quited.

    """

    def __init__(self, data=None):
        """Init CallbackQuitEvent.

        Parameters
        ----------
        data : any data type, optional
            you might use this variable to return data or values from your
            callback function to your main function, since the quited wait or
            event loop function will return this CallbackQuitEvent

        See Also
        --------
        expyriment.design.Experiment.register_wait_callback_function

        """

        self.data = data

    def __str__(self):
        return "CallbackQuitEvent: data={0}".format(self.data)


# IMPORTER FUNCTIONS

def run_py_file_command(path):
    # helper function to generate import command for extras that is Python2/3 compatible
    if PYTHON3:
        return "compile(open('{0}', 'rb').read(), '{0}', 'exec')\n".format(path)
    else:
        return "execfile(r'{0}')\n".format(path)

def get_settings_folder():
    """Return for expyriment setting folder in $HOME"""
    home = os.getenv('USERPROFILE')
    if home is None:
        if android is not None:
            home = "/storage/sdcard0/"
        else:
            home = os.getenv('HOME')
    for expy_homefolder in [".expyriment", "~expyriment"]:
        path = home + os.sep + expy_homefolder
        if os.path.isdir(path):
            return path
    return None

def get_plugins_folder():
    """Return for expyriment plugin folder in $HOME if it exists"""

    settings_folder = get_settings_folder()
    if settings_folder is None:
        return None
    path = os.path.join(settings_folder, "extras")
    if os.path.isdir(path):
        return path
    else:
        return None

def import_plugins_code(package):
    """Return the code to import all plugins from the settings folder as dict.

    Includes the module folder in $home to the path.

    Parameters
    ----------
    package : submodule name

    Returns
    -------
    out : string

    """

    code = {}
    extras_folder = get_plugins_folder()
    if extras_folder is None:
        return code
    package = "expyriment_" + package + "_extras"
    module_folder = os.path.abspath(os.path.join(extras_folder, package))

    if os.path.isdir(module_folder):
        for entry in os.listdir(module_folder):
            init_file = os.path.join(module_folder, entry, "__init__.py")
            if os.path.isfile(init_file):
                # find name of first class --> class_name
                class_name = None
                with open(init_file) as init_fl:
                    for l in init_fl:
                        if l.strip().startswith("class "):
                            l = l[(l.find("class ")+6):]
                            e = (l.find("("), l.find(":"))
                            if e[1]>1:
                                if e[0]>1:
                                    e = min(e)
                                else:
                                    e = e[1]
                                class_name = l[:e]
                                break # file loop (for..)

                if class_name is not None:
                    code[class_name] = "from {0}.{1} import {2}\n".format(package,
                                                                          entry, class_name)
    return code

def post_import_hook():
    """Execute post import file."""

    home = get_settings_folder()
    if home is None:
        return ""

    filename = home + os.sep + "post_import.py"
    if os.path.isfile(filename):
        print("process {0}".format(filename))
        return run_py_file_command(filename)
    else:
        return ""

def import_all_extras():
    """Import all extra plugins."""

    from .design import extras
    from .misc import extras
    from .io import extras
    from .stimuli import extras



#!/usr/bin/env python

"""
The Expyriment documentation

This script contains an API reference browser and search GUI interface (TK), as
well as a function to call this browser or the online documentation.

"""
from __future__ import absolute_import, print_function, division
from builtins import *
from future import standard_library
standard_library.install_aliases()

__author__ = 'Florian Krause <florian@expyriment.org> \
Oliver Lindemann <oliver@expyriment.org>'
__version__ = ''
__revision__ = ''
__date__ = ''


import os
import sys
from pydoc import getdoc as _getdoc
import inspect as _inspect
from types import ModuleType, MethodType, FunctionType

import expyriment
from ._internals import get_version


try:
    import tkinter as _tk
except:
    try:
        import tkinter as _tk  # for future (Python 3)
    except:
        _tk = None

try:
    import tkinter.ttk as _ttk
    # for OS X, if there is no Tile support
    _root = _ttk.Tk()
    _root.destroy()
except:
    _ttk = _tk  # for Python < 2.7

def _get_doc_and_function(obj):
    rtn = []
    for var in dir(obj):
        if not var.startswith("_"):
            rtn.append(var)
    return _getdoc(obj), rtn

def _read_module(mod, doc_dict):
    doc_dict[mod.__name__], classes = _get_doc_and_function(mod)
  
    p = os.path.abspath(os.path.join(os.path.split(sys.argv[0])[0], '..'))
    sys.path.insert(0, p)
   
    import expyriment
    namespace = locals()
    for cl in classes:
        cl = "{0}.{1}".format(mod.__name__, cl)
        exec("_x =" + cl, namespace)
        doc_dict[cl], functions = _get_doc_and_function(namespace['_x'])
        for fnc in functions:
            fnc = "{0}.{1}".format(cl, fnc)
            exec("_y =" + fnc, namespace)
            doc_dict[fnc], _tmp = _get_doc_and_function(namespace['_y'])

def _search_doc(search_str, doc_dict):
    """Search the documentation.

    Parameters
    ----------
    search_str : string
        string to search for (str)
    doc_dict : dict
        documentation dict to search in(dict

    """

    rtn = []
    for k in list(doc_dict.keys()):
        if k.lower().find(search_str.lower()) > -1 or\
            doc_dict[k].lower().find(search_str.lower()) > -1:
            rtn.append(k)
    return rtn

def _get_members(item_str):
    members = []
    for member in _inspect.getmembers(eval(item_str)):
        if not member[0].startswith("_"):
            members.append(item_str + "." + member[0])
    return members

def show_GUI():
    """Show the GUI."""

    from types import ModuleType
    import expyriment
    import expyriment.io.extras
    import expyriment.design.extras
    import expyriment.stimuli.extras
    import expyriment.misc.extras

    if not isinstance(_tk, ModuleType):
        raise ImportError("""API Reference Tool could not be started.
The Python package 'Tkinter' is not installed""")

    # Create the documentation dict
    doc_dict = {}
    _read_module(expyriment, doc_dict)
    _read_module(expyriment.control, doc_dict)
    _read_module(expyriment.design, doc_dict)
    _read_module(expyriment.design.extras, doc_dict)
    _read_module(expyriment.misc, doc_dict)
    _read_module(expyriment.misc.extras, doc_dict)
    _read_module(expyriment.misc.data_preprocessing, doc_dict)
    _read_module(expyriment.io, doc_dict)
    _read_module(expyriment.io.extras, doc_dict)
    _read_module(expyriment.stimuli, doc_dict)
    _read_module(expyriment.stimuli.extras, doc_dict)

    # Create e root window
    root = _tk.Tk()
    root.title("Expyriment ({0}) API Reference Tool".format(
        expyriment.__version__))
    root.minsize(996, 561)

    # Create the GUI elements
    left_frame = _ttk.Frame(root)
    search_frame = _ttk.Frame(left_frame)
    label = _ttk.Label(search_frame, text="Search:")

    search_text = _tk.StringVar()
    entry = _ttk.Entry(search_frame, textvariable=search_text, takefocus=0)
    entry.delete(0, _tk.END)

    list_frame_outer = _ttk.Frame(left_frame)
    list_frame_inner = _ttk.Frame(list_frame_outer)
    listbox = _tk.Listbox(list_frame_inner, width=40, font=("Courier", 10,
                                                           "bold"))
    scroll1 = _ttk.Scrollbar(list_frame_inner, orient=_tk.HORIZONTAL, takefocus=0)
    scroll1.config(command=listbox.xview)
    listbox.configure(xscrollcommand=scroll1.set)
    scroll2 = _ttk.Scrollbar(list_frame_outer, takefocus=0)
    scroll2.config(command=listbox.yview)
    listbox.configure(yscrollcommand=scroll2.set)

    right_frame = _ttk.Frame(root)
    text_frame_outer = _ttk.Frame(right_frame)
    text_frame_inner = _ttk.Frame(text_frame_outer)
    text = _tk.Text(text_frame_inner, width=80, background='white',
                   font=("Courier", 10), wrap=_tk.NONE, state=_tk.DISABLED,
                   takefocus=1)
    scroll3 = _ttk.Scrollbar(text_frame_inner, orient=_tk.HORIZONTAL, takefocus=0)
    scroll3.config(command=text.xview)
    text.configure(xscrollcommand=scroll3.set)
    scroll4 = _ttk.Scrollbar(text_frame_outer, takefocus=0)
    scroll4.config(command=text.yview)
    text.configure(yscrollcommand=scroll4.set)

    def update_search(_event):
        """Update the search.

        This will update the list of found items after each typed letter.

        Parameters
        ----------
        _event : dummy
            dummy argument, necessary for callbacks

        """

        global last_sel
        listbox.delete(0, _tk.END)
        value = search_text.get()
        if value == "":
            items = ["expyriment", ]
        else:
            items = _search_doc(value, doc_dict)
        for index, item in enumerate(items):
            if isinstance(eval(item), ModuleType):
                items[index] = "# " + item
            elif isinstance(eval(item), type): ##
                items[index] = "+ " + item
            elif isinstance(eval(item), MethodType):
                items[index] = "- " + item
            elif isinstance(eval(item), FunctionType):
                items[index] = "= " + item
            else:
                items[index] = "@ " + item
        items = sorted(items)
        if items == []:
            text.config(state=_tk.NORMAL)
            text.delete(1.0, _tk.END)
            text.config(state=_tk.DISABLED)
            last_sel = None
        for index, item in enumerate(items):
            listbox.insert(_tk.END, item)
            if isinstance(eval(item[2:]), (type, ModuleType)):
                listbox.itemconfig(index, fg="blue", selectforeground="blue")
        listbox.selection_set(_tk.ACTIVE)
        return True

    def poll():
        """Poll the GUI.

        This will update the documentation according to the selected item.

        """

        global last_sel
        text.after(100, poll)
        sel = listbox.curselection()
        if sel != ():
            item = listbox.get(int(sel[0]))
            if last_sel != item:
                last_sel = item
                if item == "..":
                    text.config(state=_tk.NORMAL)
                    text.delete(1.0, _tk.END)
                    text.config(state=_tk.DISABLED)
                if item != "..":
                    item = item[2:]
                    text.config(state=_tk.NORMAL)
                    text.delete(1.0, _tk.END)
                    text.tag_config("heading", font=("Courier", 12, "bold"))
                    text.insert(_tk.END, item, "heading")
                    text.insert(_tk.END, "\n\n")
                    if isinstance(eval(item), type):
                        text.insert(_tk.END, doc_dict[item])
                        definition = "".join(_inspect.getsourcelines(
                            eval(item))[0])
                        start = definition.find("def __init__(self") + 17
                        end = definition.find(")", start)
                        if definition[start] == ",":
                            call = "(" + \
                                    definition[start + 1:end].lstrip() + ")"
                        else:
                            call = "()"
                        call = call.replace(" " * 16,
                                            " " * len(item.split(".")[-1]))
                        text.insert(_tk.END, "\n\n\n\n")
                        text.tag_config("item",
                                        font=("Courier", 10, "bold"))
                        text.tag_config("call", font=("Courier", 10,
                                                      "italic"))
                        text.insert(_tk.END, item.split(".")[-1], "item")
                        text.insert(_tk.END, call, "call")
                        text.insert(_tk.END, "\n\n")
                        text.insert(_tk.END, _getdoc(
                            eval(item + "." + "__init__")))
                    elif isinstance(eval(item), FunctionType):
                        definition = "".join(_inspect.getsourcelines(
                            eval(item))[0])
                        text.tag_config("item",
                                        font=("Courier", 10, "bold"))
                        start = definition.find("(") + 1
                        end = definition.find(")", start)
                        call = "(" + definition[start:end].lstrip() + ")"
                        call = call.replace(
                            "    " + " "*len(item.split(".")[-1]) + " ",
                            " "*len(item.split(".")[-1] + " "))
                        text.tag_config("call", font=("Courier", 10, "italic"))
                        text.insert(_tk.END, item.split(".")[-1], "item")
                        text.insert(_tk.END, call, "call")
                        text.insert(_tk.END, "\n\n")
                        text.insert(_tk.END, doc_dict[item])
                    elif isinstance(eval(item), MethodType):
                        definition = "".join(_inspect.getsourcelines(
                            eval(item))[0])
                        text.tag_config("item",
                                        font=("Courier", 10, "bold"))
                        start = definition.find("(self") + 1
                        end = definition.find(")", start)
                        if definition[start + 4] == ",":
                            call = "(" + \
                                    definition[start + 5:end].lstrip() + ")"
                        else:
                            call = "()"
                        call = call.replace(
                            "        " + " "*len(item.split(".")[-1]) + " ",
                            " "*len(item.split(".")[-1]) + " ")
                        text.tag_config("call", font=("Courier", 10, "italic"))
                        text.insert(_tk.END, item.split(".")[-1], "item")
                        text.insert(_tk.END, call, "call")
                        text.insert(_tk.END, "\n\n")
                        text.insert(_tk.END, doc_dict[item])
                    elif isinstance(eval(item), (int, bytes,bool, list,
                                              tuple, dict)):
                        pass
                    else:
                        if isinstance(eval(item), property):
                            if eval(item).fset is None:
                                text.insert(_tk.END, "Read-only!")
                        else:
                            text.insert(_tk.END, doc_dict[item])

                    text.config(state=_tk.DISABLED)

    def member_list(_event, member=None):
        """Show a list of members of an item.

        This will show_GUI the list of all members (modules, classes, methods, ...)
        for a given item. If no item is given, it will take the currently
        selected one.

        Parameters
        ----------
        _event : dummy
            dummy argument, necessary for callbacks
        member : string, optional
            item to show_GUI members for

        """

        global last_item
        if member is not None:
            item = member
        else:
            try:
                item = listbox.get(int(listbox.curselection()[0]))
                if item == "..":
                    tmp = item = last_item
                else:
                    item = item[2:]
                    tmp = item
                if item == "":
                    update_search(None)
                else:
                    if isinstance(eval(item), (type, ModuleType)):
                        s = tmp.split(".")
                        if len(s) >= 1:
                            last_item = ".".join(s[0:-1])
                        else:
                            last_item = None
                        items = _get_members(item)
                        items.insert(0, "..")
                        entry.delete(0, _tk.END)
                        listbox.delete(0, _tk.END)
                        for index, item in enumerate(items):
                            if item != "..":
                                if isinstance(eval(item), ModuleType):
                                    items[index] = "# " + item
                                elif isinstance(eval(item), type):
                                    items[index] = "+ " + item
                                elif isinstance(eval(item), MethodType):
                                    items[index] = "- " + item
                                elif isinstance(eval(item), FunctionType):
                                    items[index] = "= " + item
                                else:
                                    items[index] = "@ " + item
                        tmp = items
                        items = []
                        items.append(tmp[0])
                        items.extend(sorted(tmp[1:]))
                        for index, item in enumerate(items):
                            listbox.insert(_tk.END, item)
                            if item == ".." or \
                                    isinstance(eval(item[2:]), (type, ModuleType)):
                                listbox.itemconfig(index, fg="blue",
                                                   selectforeground="blue")
                        listbox.selection_set(0)
            except:
                pass

    # Position the GUI elements
    left_frame.pack(side=_tk.LEFT, expand=1, fill=_tk.BOTH)
    search_frame.pack(side=_tk.TOP, fill=_tk.BOTH)
    label.pack(side=_tk.LEFT)
    entry.pack(side=_tk.LEFT, expand=1, fill=_tk.X)
    list_frame_outer.pack(side=_tk.BOTTOM, expand=1, fill=_tk.BOTH)
    list_frame_inner.pack(side=_tk.LEFT, expand=1, fill=_tk.BOTH)
    listbox.pack(side=_tk.TOP, expand=1, fill=_tk.BOTH)
    scroll1.pack(side=_tk.BOTTOM, fill=_tk.X)
    scroll2.pack(side=_tk.LEFT, fill=_tk.Y)
    right_frame.pack(side=_tk.RIGHT, expand=2, fill=_tk.BOTH)
    text_frame_outer.pack(side=_tk.BOTTOM, expand=2, fill=_tk.BOTH)
    text_frame_inner.pack(side=_tk.LEFT, expand=2, fill=_tk.BOTH)
    text.pack(side=_tk.TOP, expand=2, fill=_tk.BOTH)
    scroll3.pack(side=_tk.BOTTOM, fill=_tk.X)
    scroll4.pack(side=_tk.LEFT, fill=_tk.Y)

    # Create Keybindings
    root.bind("<Control-f>", lambda x: entry.focus())
    root.bind("<Control-q>", lambda x: root.quit())
    root.bind("<F1>", lambda x: show_help())
    listbox.bind("<Double-Button-1>", member_list)
    listbox.bind("<Return>", member_list)
    entry.bind("<KeyRelease>", update_search)

    def show_about():
        """Show the about dialogue window"""

        aboutdialogue = _tk.Toplevel(root, padx=5, pady=5)
        aboutdialogue.title("About")
        aboutdialogue.transient(root)
        aboutdialogue.grab_set()
        aboutdialogue.focus_set()
        aboutdialogue.bind("<Button-1>", lambda x: aboutdialogue.destroy())
        aboutdialogue.bind("<Escape>", lambda x: aboutdialogue.destroy())
        aboutdialogue.bind("<Return>", lambda x: aboutdialogue.destroy())
        aboutlabel1 = _ttk.Label(aboutdialogue,
                               text="Expyriment API Reference Tool",
                               font=("Arial", "15", "bold"))
        aboutlabel1.pack()
        aboutlabel2 = _ttk.Label(aboutdialogue,
                               text="Expyriment {0}".format(
                                   expyriment.get_version()),
                               font=("Arial", "8", "italic"))
        aboutlabel2.pack()
        aboutlabel3 = _ttk.Label(aboutdialogue,
                               text="",
                               font=("Arial", "11"))
        aboutlabel3.pack()
        aboutlabel4 = _ttk.Label(aboutdialogue,
                        text="Florian Krause <florian@expyriment.org>",
                               font=("Arial", "9"))
        aboutlabel4.pack()
        aboutlabel5 = _ttk.Label(aboutdialogue,
                        text="Oliver Lindemann <oliver@expyriment.org>",
                               font=("Arial", "9"))
        aboutlabel5.pack()

    def show_help():
        """Show the help dialogue window"""

        helpdialogue = _tk.Toplevel(root, width=200, height=300, padx=5,
                                   pady=5)
        helpdialogue.title("Help Contents")
        helpdialogue.transient(root)
        helpdialogue.grab_set()
        helpdialogue.bind("<Escape>", lambda x: helpdialogue.destroy())
        helpdialogue.bind("<Return>", lambda x: helpdialogue.destroy())

        helpframe = _ttk.Frame(helpdialogue)
        helpframe.pack()
        helptextbox = _tk.Text(helpframe, highlightthickness=0, wrap=_tk.WORD,
                               font=("Courier", 10))
        documentation = """The Expyriment API Reference Tool allows you to browse and search the Expyriment API.

Double clicking on blue coloured items will show their content.
Double clicking on ".." will go up one level.

The search is instant, which means every typed letter will update the results immediately.

The following symbols are used to denote the type of an item:

    # Module
    + Class
    - Method
    = Function
    @ Attribute"""
        helptextbox.insert('1.0', documentation)
        helptextbox.pack(side=_tk.LEFT)
        helptextbox.focus_set()
        scrollbar = _ttk.Scrollbar(helpframe, takefocus=False,
                                 command=helptextbox.yview)
        scrollbar.pack(side=_tk.RIGHT, fill=_tk.Y)
        helptextbox.config(state=_tk.DISABLED,
                           yscrollcommand=scrollbar.set)
        closebutton = _ttk.Button(helpdialogue, text="Close",
                                takefocus=_tk.FALSE,
                                command=helpdialogue.destroy)
        closebutton.pack()

    # Create a menu
    menubar = _tk.Menu(root)
    filemenu = _tk.Menu(menubar, tearoff=0)
    filemenu.add_command(label="Quit", command=root.quit, accelerator="Ctrl+Q")
    menubar.add_cascade(label="File", menu=filemenu)
    helpmenu = _tk.Menu(menubar, tearoff=0)
    helpmenu.add_command(label="Contents", command=show_help, accelerator="F1")
    helpmenu.add_command(label="About", command=show_about)
    menubar.add_cascade(label="Help", menu=helpmenu)

    root.config(menu=menubar)

    # Start the main loop
    global last_item
    last_item = None
    update_search(None)

    global last_sel
    last_sel = None
    poll()

    entry.focus()

    _tk.mainloop()


def show_documentation(docu_type=None):
    """Show the Expyriment documentation.


    Parameters
    ----------
    docu_type : int
        the documentation type:
        1 = open online documentation
        2 = open online API reference
        3 = open API reference and search tool

    """

    def call_info():
        print("")
        print("Call show_documentation with the following arguments to get further information:")
        print("     show_documentation(1) -- Open local documentation in web browser")
        print("     show_documentation(2) -- Open online documentation in web browser")
        print("     show_documentation(3) -- Open API Reference Tool")
        print("")

    import subprocess
    import os
    import sys
    import webbrowser

    f = os.path.abspath(__file__)
    path = os.path.abspath(os.path.join(os.path.split(f)[0], ".."))
    if docu_type is None:
        print("Welcome to Expyriment {0}".format(get_version()))
        print("")
        author = __author__.replace(",", ",\n        ")
        print("Website: http://expyriment.org")
        print("License: GNU GPL v3")
        print("Authors: {0}".format(author))
        call_info()
    elif docu_type == 1:
        docu = os.path.join(sys.prefix, 'share', 'expyriment', 'documentation', 'html', 'index.html')
        docu_fallback = os.path.join('usr', 'local', 'share', 'expyriment', 'documentation', 'html', 'index.html')
        if os.path.exists(docu):
            webbrowser.open('file://' + docu,
                new=1)
        elif os.path.exists(docu_fallback):
            webbrowser.open('file://' + docu_fallback,
                new=1)
        else:
            print("No local documentation found")
    elif docu_type == 2:
        webbrowser.open(
            "http://docs.expyriment.org/",
            new=1)
    elif docu_type == 3:
        python_executable = sys.executable.replace("pythonw.exe", "python.exe")
        call = '"' + "{0}".format(python_executable) + \
                '" -m expyriment._api_reference_tool'
        _proc = subprocess.Popen(
            call,
            shell=True,
            stdin=None,
            stdout=None,
            cwd=path)
    else:
        print("Unknown documentation type")
        call_info()

if __name__ == "__main__":
    show_GUI()



"""
Default settings for the stimuli package.

This module contains default values for all optional arguments in the init
function of all classes in this package.

"""

__author__ = 'Florian Krause <florian@expyriment.org, \
Oliver Lindemann <oliver@expyriment.org>'
__version__ = ''
__revision__ = ''
__date__ = ''


from tempfile import mkdtemp as _mkdtemp
from shutil import rmtree as _rmtree
import atexit as _atexit

# Visual
visual_position = (0, 0)

# Canvas
canvas_colour = None  # 'None' is transparent
canvas_position = (0, 0)

# TextLine
textline_text_font = None  # 'None' is experiment_text_font
textline_text_size = None  # 'None' is experiment_text_size
textline_text_bold = False
textline_text_italic = False
textline_text_underline = False
textline_text_colour = None  # 'None' is experiment_text_colour
textline_background_colour = None  # 'None' is transparent
textline_position = (0, 0)

# TextBox
textbox_text_font = None  # 'None' is experiment_text_font
textbox_text_size = None  # 'None' is experiment_text_size
textbox_text_bold = False
textbox_text_italic = False
textbox_text_underline = False
textbox_text_justification = 1
textbox_text_colour = None  # 'None' is experiment_text_colour
textbox_background_colour = None  # 'None' is transparent
textbox_position = (0, 0)
textbox_do_not_trim_words = False

# TextScreen
textscreen_heading_font = None  # 'None' is experiment_heading_font
textscreen_heading_size = None  # 'None' is experiment_heading_size
textscreen_heading_bold = False
textscreen_heading_italic = False
textscreen_heading_underline = False
textscreen_heading_colour = None  # 'None' is experiment_heading_colour
textscreen_text_font = None  # 'None' is experiment_text_font
textscreen_text_size = None  # 'None' is experiment_text_size
textscreen_text_bold = False
textscreen_text_italic = False
textscreen_text_underline = False
textscreen_text_colour = None  # 'None' is experiment_text_colour
textscreen_text_justification = 1
textscreen_background_colour = None  # 'None' is transparent
textscreen_size = None  # 'None' is 4/5 of full screen
textscreen_position = (0, 0)

# Ellipse
ellipse_colour = None  # 'None' is experiment_text_colour
ellipse_line_width = 0
ellipse_position = (0, 0)
ellipse_anti_aliasing = 0

# FixCross
fixcross_colour = None  # 'None' is experiment_text_colour
fixcross_size = (20, 20)
fixcross_line_width = 2
fixcross_position = (0, 0)
fixcross_anti_aliasing = 0

# Circle
circle_colour = None  # 'None' is experiment_text_colour
circle_position = (0, 0)
circle_anti_aliasing = 0
circle_line_width = 0

# Shape
shape_colour = None  # 'None' is experiment_text_colour
shape_position = (0, 0)
shape_anti_aliasing = 0

# Line
line_colour = None  # 'None' is experiment_text_colour
line_anti_aliasing = 0

# Rectangle
rectangle_colour = None  # 'None' is experiment_text_colour
rectangle_position = (0, 0)
rectangle_line_width = 0
rectangle_corner_rounding = 0
rectangle_corner_anti_aliasing = 0

# Picture
picture_position = (0, 0)

# Video
video_position = [0, 0]
video_backend = "mediadecoder"  # 'mediadecoder' or 'pygame'

# Tone
tone_frequency = 440
tone_samplerate = 44100
tone_bitdepth = 16
tone_amplitude = 0.5

# Create tmp for compressed stimuli folder
try:
    tempdir = _mkdtemp(prefix="expyriment_")
except:
    tempdir = None
    pass

def _remove_tempdir():
    global tempdir
    try:
        _rmtree(tempdir)
    except:
        pass

_atexit.register(_remove_tempdir)



"""
Keyboard input.

This module contains a class implementing pygame keyboard input.

"""
from __future__ import absolute_import, print_function, division
from builtins import *

__author__ = 'Florian Krause <florian@expyriment.org>, \
Oliver Lindemann <oliver@expyriment.org>'
__version__ = ''
__revision__ = ''
__date__ = ''


import time, sys
from types import FunctionType

import pygame

try:
    import android.show_keyboard as android_show_keyboard
    import android.hide_keyboard as android_hide_keyboard
except ImportError:
    android_show_keyboard = android_hide_keyboard = None

from . import defaults

from ..misc._timer import get_time
from ..misc import unicode2byte
from  ._input_output import Input
from .. import  _internals

quit_key = None
pause_key = None
end_function = None
pause_function = None

class Keyboard(Input):
    """A class implementing a keyboard input.

    Calling `expyriment.control.initialize(exp)` will automatically create a
    keyboard instance and will reference it in exp.keyboard for easy access.

    """

    @staticmethod
    def process_control_keys(key_event=None, quit_confirmed_function=None):
        """Check if quit_key or pause_key has been pressed.

        Reads pygame event cue if no key_event is specified.

        Parameters
        ----------
        key_event : int, optional
            key event to check. If not defined, the Pygame event queue will be
            checked for key down events.
        quit_confirmed_function : function, optional
            function to be called if quitting has been confirmed.

        Returns
        -------
        out : bool
            True if quitting or pause screen has been displayed,
            False otherwise

        """

        if key_event:
            if key_event.type == pygame.KEYDOWN:
                if key_event.key == quit_key and \
                   end_function is not None:
                    confirm = end_function(
                        confirmation=True,
                        pre_quit_function=quit_confirmed_function)
                    if confirm:
                        sys.exit()
                    return True
                elif key_event.key == pause_key and \
                        pause_function is not None:
                    pause_function()
                    return True
        else:
            ## -------
            for event in pygame.event.get(pygame.KEYDOWN):
                # recursion
                return Keyboard.process_control_keys(event,
                                                     quit_confirmed_function)

            # code above is as in 0.9. However, it only processes to first key
            # press and deletes all other keydown events. This is a bug,
            # but remains unchanged in the current version to keep event
            # polling in 0.10 identical to 0.9
            #
            # TODO for 1.0: Process all keydown events
            # ...
            #             for event in pygame.event.get(pygame.KEYDOWN):
            #                 if Keyboard.process_control_keys(event,
            #                                                  quit_confirmed_function):
            #                     return True
            #             return False
            # ...
            ## -------

        return False


    def __init__(self, default_keys=None):
        """Create a keyboard input.

        Parameters
        ----------
        default_keys : int or list, optional
            a default key or list of default keys

        """

        if default_keys is not None:
            self._default_keys = default_keys
        else:
            self._default_keys = defaults.keyboard_default_keys
        Input.__init__(self)

    @property
    def default_keys(self):
        """Getter for default keys"""

        return self._default_keys

    @default_keys.setter
    def default_keys(self, value):
        """Setter for default keys"""

        self._default_keys = value

    @staticmethod
    def get_quit_key():
        """Returns the currently defined quit key """

        return quit_key

    @staticmethod
    def get_pause_key():
        """Returns the currently defined pause key"""

        return pause_key

    @staticmethod
    def set_quit_key(value):
        """Set the currently defined quit key"""

        global quit_key
        quit_key = value

    @staticmethod
    def set_pause_key(value):
        """Set the currently defined pause key"""

        global pause_key
        pause_key = value

    def clear(self):
        """Clear the event queue from keyboard events."""

        pygame.event.clear(pygame.KEYDOWN)
        pygame.event.clear(pygame.KEYUP)

        if self._logging:
            _internals.active_exp._event_file_log("Keyboard,cleared", 2)

    def read_out_buffered_keys(self):
        """Reads out all keydown events and clears queue."""

        pygame.event.pump()
        pygame.event.clear(pygame.KEYUP)
        rtn = []
        for event in pygame.event.get(pygame.KEYDOWN):
            Keyboard.process_control_keys(event)
            rtn.append(event.key)
        return rtn

    def check(self, keys=None, check_for_keyup=False,
              check_for_control_keys=True):
        """Check if keypress is in event queue.

        Parameters
        ----------
        keys : int or list, optional
            a specific key or list of keys to check
        check_for_keyup : bool, optional
            if True checks for key-up (default = False)
        check_for_control_keys : bool, optional
            checks if control key has been pressed (default = True)

        Returns
        -------
        key : int
            pressed key or None. Only the first occurrence is returned!

        Notes
        -----
        Keys are defined by keyboard constants (please see misc.constants).

        When checking for key-down (default), all key-up events are first
        cleared, and when checking for key-up events, all key-down events are
        first cleared!

        """

        # TODO for 1.0: Check should not return None. Think about the
        #               introduction of QuitAttemptEvent!

        if keys is None:
            keys = self.default_keys
        else:
            try:
                keys = list(keys)
            except:
                keys = [keys]

        pygame.event.pump()

        if check_for_keyup:
            target_event = pygame.KEYUP
            if check_for_control_keys:
                self.check(keys=None, check_for_keyup=False,
                           check_for_control_keys=True)
                check_for_control_keys = False
        else:
            target_event = pygame.KEYDOWN
            pygame.event.clear(pygame.KEYUP)

        for event in pygame.event.get(target_event):
            if check_for_control_keys:
                Keyboard.process_control_keys(event)
            if keys:
                if event.key in keys:
                    if self._logging:
                        _internals.active_exp._event_file_log(
                            "Keyboard,received,{0},check".format(event.key))
                    return event.key
            else:
                if self._logging:
                    _internals.active_exp._event_file_log(
                        "Keyboard,received,{0},check".format(event.key), 2)
                return event.key
        return None

    def wait(self, keys=None, duration=None, wait_for_keyup=False,
             callback_function=None, process_control_events=True):
        """Wait for keypress(es) (optionally for a certain amount of time).

        This function will wait for a keypress and returns the found key as
        well as the reaction time.
        (This function clears the event queue!)

        Parameters
        ----------
        keys : int or list, optional
            a specific key or list of keys to wait for
        duration : int, optional
            maximal time to wait in ms
        wait_for_keyup : bool, optional
            if True it waits for key-up (default=False)
        callback_function : function, optional
            function to repeatedly execute during waiting loop
        process_control_events : bool, optional
            process ``io.Keyboard.process_control_keys()`` and
            ``io.Mouse.process_quit_event()`` (default = True)

        Returns
        -------
        found : char
            pressed character
        rt : int
            reaction time in ms

        Notes
        -----
        Keys are defined by keyboard constants (please see misc.constants).

        See Also
        --------
        expyriment.design.Experiment.register_wait_callback_function

        """

        if _internals.skip_wait_methods:
            return None, None
        if android_show_keyboard is not None:
            android_show_keyboard()
        start = get_time()
        rt = None
        found_key = None
        self.clear()
        if keys is None:
            keys = self.default_keys
        else:
            try:
                keys = list(keys)
            except:
                keys = [keys]
        if wait_for_keyup:
            target_event = pygame.KEYUP
        else:
            target_event = pygame.KEYDOWN
        pygame.event.pump()
        done = False
        while not done:
            if isinstance(callback_function, FunctionType):
                rtn_callback = callback_function()
                if isinstance(rtn_callback, _internals.CallbackQuitEvent):
                    done = True
                    found_key = rtn_callback
                    rt = int((get_time() - start) * 1000)
            if _internals.active_exp.is_initialized:
                rtn_callback = _internals.active_exp._execute_wait_callback()
                if isinstance(rtn_callback, _internals.CallbackQuitEvent):
                    done = True
                    found_key = rtn_callback
                    rt = int((get_time() - start) * 1000)
                if process_control_events and \
                    _internals.active_exp.mouse.process_quit_event():
                    done = True
            for event in pygame.event.get([pygame.KEYDOWN, pygame.KEYUP]):
                if _internals.active_exp.is_initialized and \
                   process_control_events and \
                   Keyboard.process_control_keys(event):
                    done = True
                elif event.type == target_event:
                    if keys is not None:
                        if event.key in keys:
                            rt = int((get_time() - start) * 1000)
                            found_key = event.key
                            done = True
                    else:
                        rt = int((get_time() - start) * 1000)
                        found_key = event.key
                        done = True
            if duration and not done:
                done = int((get_time() - start) * 1000) >= duration
            time.sleep(0.0005)
        if self._logging:
            _internals.active_exp._event_file_log("Keyboard,received,{0},wait"\
                                              .format(found_key))
        if android_hide_keyboard is not None:
            android_hide_keyboard()
        return found_key, rt

    def wait_char(self, char, duration=None, callback_function=None,
                  process_control_events=True):
        """Wait for character(s) (optionally for a certain amount of time).

        This function will wait for one or more characters and returns the
        found character as well as the reaction time.
        (This function clears the event queue!)

        Parameters
        ----------
        char : int or list
            a specific character or list of characters to wait for
        duration : int, optional
            maximal time to wait in ms
        callback_function : function, optional
            function to repeatedly execute during waiting loop
        process_control_events : bool, optional
            process ``io.Keyboard.process_control_keys()`` and
            ``io.Mouse.process_quit_event()`` (default = True)

        Returns
        -------
        found : char
            pressed character
        rt : int
            reaction time in ms

        See Also
        --------
        expyriment.design.Experiment.register_wait_callback_function

        """

        if _internals.skip_wait_methods:
            return None, None
        start = get_time()
        rt = None
        found_char = None
        self.clear()
        try:
            char = list(char)
        except:
            char = [char]
        pygame.event.pump()
        done = False

        while not done:
            if isinstance(callback_function, FunctionType):
                rtn_callback = callback_function()
                if isinstance(rtn_callback, _internals.CallbackQuitEvent):
                    done = True
                    rt = int((get_time() - start) * 1000)
                    found_char = rtn_callback
            if _internals.active_exp.is_initialized:
                rtn_callback = _internals.active_exp._execute_wait_callback()
                if isinstance(rtn_callback, _internals.CallbackQuitEvent):
                    done = True
                    rt = int((get_time() - start) * 1000)
                    found_char = rtn_callback
                if process_control_events and  \
                        _internals.active_exp.mouse.process_quit_event():
                    done = True
            for event in pygame.event.get([pygame.KEYUP, pygame.KEYDOWN]):
                if _internals.active_exp.is_initialized and \
                   process_control_events and \
                   Keyboard.process_control_keys(event):
                    done = True
                elif event.type == pygame.KEYDOWN:
                    if event.unicode in char:
                        rt = int((get_time() - start) * 1000)
                        found_char = event.unicode
                        done = True
            if duration and not done:
                done = int((get_time() - start) * 1000) >= duration
            time.sleep(0.0005)
        if self._logging:
            if found_char is not None:
                _internals.active_exp._event_file_log(
                    u"Keyboard,received,{0},wait_char".format(
                        found_char))
            else:
                _internals.active_exp._event_file_log(
                    "Keyboard,received,None,wait_char")
        return found_char, rt



"""
Default settings for the design package.

This module contains default values for all optional arguments in the init
function of all classes in this package.

"""

__author__ = 'Florian Krause <florian@expyriment.org, \
Oliver Lindemann <oliver@expyriment.org>'
__version__ = ''
__revision__ = ''
__date__ = ''


# Experiment
experiment_name = None  # Set None if experiment default name should be the
#                        name of the python main file


experiment_background_colour = (0, 0, 0)
experiment_foreground_colour = (150, 150, 150)
#experiment_text_font = _str2unicode(_os.path.abspath(
#    _os.path.join(_os.path.dirname(__file__),
#                  "..", "_fonts", "FreeSans.ttf")))
experiment_text_font = "FreeSans"
experiment_text_size = 20
experiment_filename_suffix = None

# Block
block_name = 'unnamed'
max_shuffle_time = 5000

# trial_list
trial_list_directory = 'trials'



#!/usr/bin/env python

"""
A blank screen stimulus.

This module contains a class implementing a blank screen stimulus.

"""
from __future__ import absolute_import, print_function, division
from builtins import *

__author__ = 'Florian Krause <florian@expyriment.org>, \
Oliver Lindemann <oliver@expyriment.org>'
__version__ = ''
__revision__ = ''
__date__ = ''


from ._canvas import Canvas
from .. import _internals

class BlankScreen(Canvas):
    """A class implementing a blank screen."""

    def __init__(self, colour=None):
        """Create a blank screen.

        Parameters
        ----------
        colour : (int,int,int), optional
            colour of the blank screen
        """

        if colour is not None:
            self._colour = colour
        else:
            self._colour = _internals.active_exp.background_colour
        try:
            size = _internals.active_exp.screen.surface.get_size()
        except:
            raise RuntimeError("Could not get size of screen!")
        Canvas.__init__(self, size, colour=self._colour)


    @staticmethod
    def _demo(exp=None):
        if exp is None:
            from .. import control
            control.set_develop_mode(True)
            control.defaults.event_logging.event_logging = 0
            exp_ = control.initialize()
        blankscreen = BlankScreen()
        blankscreen.present()
        if exp is None:
            exp_.clock.wait(1000)



#!/usr/bin/env python

"""
A text screen stimulus.

This module contains a class implementing a text screen stimulus.

"""
from __future__ import absolute_import, print_function, division
from builtins import (ascii, bytes, chr, dict, filter, hex, input,
                      int, map, next, oct, pow, range, round,
                      str, super, zip)  # without open, because
                        # pygame.font.Font needs old file object under PY2


__author__ = 'Florian Krause <florian@expyriment.org>, \
Oliver Lindemann <oliver@expyriment.org>'
__version__ = ''
__revision__ = ''
__date__ = ''


import pygame

from . import defaults
from ._stimulus import Stimulus
from ._visual import Visual
from ._textline import TextLine
from ._textbox import TextBox
from ..misc import find_font, unicode2byte
from .. import _internals


class TextScreen(Visual):
    """A class implementing a screen with heading and text."""

    def __init__(self, heading, text, position=None, heading_font=None,
                 heading_size=None, heading_bold=None, heading_italic=None,
                 heading_underline=None, heading_colour=None, text_font=None,
                 text_size=None, text_bold=None, text_italic=None,
                 text_underline=None, text_colour=None,
                 text_justification=None, background_colour=None, size=None):
        """Create a text screen.

        Parameters
        ----------
        heading : str
            heading of the text screen
        text : str
            text of the text screen
        position : (int, int), optional
            position of the stimulus
        heading_font : str, optional
            heading font to use
        heading_size : int, optional
            heading font size
        heading_bold : bool, optional
            heading should be bold
        heading_italic : bool, optional
            heading should be italic
        heading_underline : bool, optional
            heading should get an underline
        heading_colour : (int,int,int), optional
            heding colour
        text_font : str, optional
            text font to use
        text_size : int, optional
            text font size
        text_bold : bool, optional
            text should be bold
        text_italic : bool, optional
            text should be italic
        text_underline : bool, optional
            text should get an underline
        text_colour : (int,int,int), optional
            text colour
        text_justification : int, optional
            0 (Left), 1(center), 2(right) (int) (optional)
        background_colour : (int, int, int), optional
            background_colour
        size : (int, int), optional
            size of the text screen

        """

        if position is None:
            position = defaults.textscreen_position
        Visual.__init__(self, position, log_comment="text_screen")
        self._heading = heading
        self._text = text
        if heading_font is None:
            heading_font = defaults.textscreen_heading_font
        if heading_font is None:
            heading_font = _internals.active_exp.text_font
        if heading_font is None:
            heading_font = "FreeSans"
        self._heading_font = find_font(heading_font)
        try:
            with open(self._heading_font, 'rb') as f:
                pygame.font.Font(f, 10)
        except:
            raise IOError("Font '{0}' not found!".format(heading_font))
        if heading_size is None:
            heading_size = defaults.textscreen_heading_size
        if heading_size:
            self._heading_size = heading_size
        else:
            self._heading_size = int(_internals.active_exp.text_size
                                     * 1.2)
        if heading_bold is not None:
            self._heading_bold = heading_bold
        else:
            self._heading_bold = defaults.textscreen_heading_bold
        if heading_italic is not None:
            self._heading_italic = heading_italic
        else:
            self._heading_italic = \
                defaults.textscreen_heading_italic
        if heading_underline is not None:
            self._heading_underline = heading_underline
        else:
            self._heading_underline = \
                defaults.textscreen_heading_underline
        if heading_colour is None:
            heading_colour = defaults.textscreen_heading_colour
        if heading_colour is not None:
            self._heading_colour = heading_colour
        else:
            self._heading_colour = _internals.active_exp.foreground_colour
        if text_font is None:
            text_font = defaults.textscreen_text_font
        if text_font is not None:
            self._text_font = find_font(text_font)
        else:
            self._text_font = find_font(_internals.active_exp.text_font)
        try:
            with open(self._text_font, 'rb') as f:
                pygame.font.Font(f, 10)
        except:
            raise IOError("Font '{0}' not found!".format(text_font))
        if text_size is None:
            self._text_size = defaults.textscreen_text_size
        if text_size is not None:
            self._text_size = text_size
        else:
            self._text_size = _internals.active_exp.text_size
        if text_bold is not None:
            self._text_bold = text_bold
        else:
            self._text_bold = defaults.textscreen_text_bold
        if text_italic is not None:
            self._text_italic = text_italic
        else:
            self._text_italic = defaults.textscreen_text_italic
        if text_underline is not None:
            self._text_underline = text_underline
        else:
            self._text_underline = defaults.textscreen_text_underline
        if text_colour is None:
            text_colour = defaults.textscreen_text_colour
        if text_colour is not None:
            self._text_colour = text_colour
        else:
            self._text_colour = _internals.active_exp.foreground_colour
        if text_justification is not None:
            self._text_justification = text_justification
        else:
            self._text_justification = \
                defaults.textscreen_text_justification
        if size is not None:
            self._size = size
        else:
            size = defaults.textscreen_size
            if size is None:
                try:
                    self._size = (
                        _internals.active_exp.screen.surface.get_size()[0] -
                        _internals.active_exp.screen.surface.get_size()[0]
                        // 5,
                        _internals.active_exp.screen.surface.get_size()[1] -
                        _internals.active_exp.screen.surface.get_size()[1]
                        // 5)
                except:
                    raise RuntimeError("Cannot get size of screen!")

        if background_colour is not None:
            self._background_colour = background_colour
        else:
            self._background_colour = \
                defaults.textscreen_background_colour

    _getter_exception_message = "Cannot set {0} if surface exists!"

    @property
    def heading(self):
        """Getter for heading."""

        return self._heading

    @heading.setter
    def heading(self, value):
        """Setter for heading."""

        if self.has_surface:
            raise AttributeError(TextScreen._getter_exception_message.format(
                "heading"))
        else:
            self._heading = value

    @property
    def text(self):
        """Getter for text."""

        return self._text

    @text.setter
    def text(self, value):
        """Setter for text."""

        if self.has_surface:
            raise AttributeError(TextScreen._getter_exception_message.format(
                "text"))
        else:
            self._text = value

    @property
    def text_font(self):
        """Getter for text_font."""

        return self._text_font

    @text_font.setter
    def text_font(self, value):
        """Setter for text_font."""

        if self.has_surface:
            raise AttributeError(TextScreen._getter_exception_message.format(
                "text_font"))
        else:
            self._text_font = value

    @property
    def text_size(self):
        """Getter for text_size."""

        return self._text_size

    @text_size.setter
    def text_size(self, value):
        """Setter for text_size."""

        if self.has_surface:
            raise AttributeError(TextScreen._getter_exception_message.format(
                "text_size"))
        else:
            self._text_size = value

    @property
    def text_bold(self):
        """Getter for text_bold."""

        return self._text_bold

    @text_bold.setter
    def text_bold(self, value):
        """Setter for text_bold."""

        if self.has_surface:
            raise AttributeError(TextScreen._getter_exception_message.format(
                "text_bold"))
        else:
            self._text_bold = value

    @property
    def text_italic(self):
        """Getter for text_italic."""

        return self._text_italic

    @text_italic.setter
    def text_italic(self, value):
        """Setter for text_italic."""

        if self.has_surface:
            raise AttributeError(TextScreen._getter_exception_message.format(
                "text_italic"))
        else:
            self._text_italic = value

    @property
    def text_underline(self):
        """Getter for text_underline."""

        return self._text_underline

    @text_underline.setter
    def text_underline(self, value):
        """Setter for text_underline."""

        if self.has_surface:
            raise AttributeError(TextScreen._getter_exception_message.format(
                "text_underline"))
        else:
            self._text_underline = value

    @property
    def text_colour(self):
        """Getter for text_colour."""

        return self._text_colour

    @text_colour.setter
    def text_colour(self, value):
        """Setter for text_colour."""

        if self.has_surface:
            raise AttributeError(TextScreen._getter_exception_message.format(
                "text_colour"))
        else:
            self._text_colour = value

    @property
    def heading_font(self):
        """Getter for heading_font."""

        return self._heading_font

    @heading_font.setter
    def heading_font(self, value):
        """Setter for heading_font."""

        if self.has_surface:
            raise AttributeError(TextScreen._getter_exception_message.format(
                "heading_font"))
        else:
            self._heading_font = value

    @property
    def heading_size(self):
        """Getter for heading_size."""

        return self._heading_size

    @heading_size.setter
    def heading_size(self, value):
        """Setter for heading_size."""

        if self.has_surface:
            raise AttributeError(TextScreen._getter_exception_message.format(
                "heading_size"))
        else:
            self._heading_size = value

    @property
    def heading_bold(self):
        """Getter for heading_bold."""

        return self._heading_bold

    @heading_bold.setter
    def heading_bold(self, value):
        """Setter for heading_bold."""

        if self.has_surface:
            raise AttributeError(TextScreen._getter_exception_message.format(
                "heading_bold"))
        else:
            self._heading_bold = value

    @property
    def heading_italic(self):
        """Getter for heading_italic."""

        return self._heading_italic

    @heading_italic.setter
    def heading_italic(self, value):
        """Setter for heading_italic."""

        if self.has_surface:
            raise AttributeError(TextScreen._getter_exception_message.format(
                "heading_italic"))
        else:
            self._heading_italic = value

    @property
    def heading_underline(self):
        """Getter for heading_underline."""

        return self._heading_underline

    @heading_underline.setter
    def heading_underline(self, value):
        """Setter for heading_underline."""

        if self.has_surface:
            raise AttributeError(TextScreen._getter_exception_message.format(
                "heading_underline"))
        else:
            self._heading_underline = value

    @property
    def heading_colour(self):
        """Getter for heading_colour."""

        return self._heading_colour

    @heading_colour.setter
    def heading_colour(self, value):
        """Setter for heading_colour."""

        if self.has_surface:
            raise AttributeError(TextScreen._getter_exception_message.format(
                "heading_colour"))
        else:
            self._heading_colour = value

    @property
    def background_colour(self):
        """Getter for background_colour."""

        return self._background_colour

    @background_colour.setter
    def background_colour(self, value):
        """Setter for background_colour."""

        if self.has_surface:
            raise AttributeError(TextScreen._getter_exception_message.format(
                "background_colour"))
        else:
            self._background_colour = value

    @property
    def size(self):
        """Getter for size."""

        return self._size

    @size.setter
    def size(self, value):
        """Setter for size."""

        if self.has_surface:
            raise AttributeError(TextScreen._getter_exception_message.format(
                "size"))
        else:
            self._size = value

    @property
    def text_justification(self):
        """Getter for text_justification."""

        return self._text_justification

    @text_justification.setter
    def text_justification(self, value):
        """Setter for text_justification."""

        if self.has_surface:
            raise AttributeError(TextScreen._getter_exception_message.format(
                "text_justification"))
        else:
            self._text_justification = value

    def _create_surface(self):
        """Create the surface of the stimulus."""

        surface = pygame.surface.Surface(self.size,
                                         pygame.SRCALPHA).convert_alpha()
        if self.background_colour is not None:
            surface.fill(self.background_colour)
        header = TextLine(text=self.heading, text_size=self.heading_size,
                          text_colour=self.heading_colour,
                          background_colour=self.background_colour,
                          text_font=self.heading_font,
                          text_bold=self.heading_bold,
                          text_italic=self.heading_italic,
                          text_underline=self.heading_underline)
        Stimulus._id_counter -= 1
        box = TextBox(text=self.text, text_font=self.text_font,
                      text_size=self.text_size, text_bold=self.text_bold,
                      text_italic=self.text_italic,
                      text_underline=self.text_underline,
                      text_colour=self.text_colour,
                      background_colour=self.background_colour,
                      size=(self.size[0], self.size[1] - self.size[1] // 5),
                      text_justification=self.text_justification)
        Stimulus._id_counter -= 1
        surface.blit(header._get_surface(),
                     (self.size[0] // 2 - header.surface_size[0] // 2,
                      0))
        surface.blit(box._get_surface(),
                     (self.size[0] // 2 - box.size[0] // 2, self.size[1] // 5))
        return surface

    @staticmethod
    def _demo(exp=None):
        if exp is None:
            from .. import control
            control.set_develop_mode(True)
            control.defaults.event_logging = 0
            exp_ = control.initialize()
        textscreen = TextScreen("Hello World",
                                "Line one.\nLine two.\nLine three.")
        textscreen.present()
        if exp is None:
            exp_.clock.wait(1000)



"""
The permute module.

This module implements permutation of blocks, trials and conditions.

"""
from __future__ import absolute_import, print_function, division
from builtins import *

__author__ = 'Florian Krause <florian@expyriment.org>, \
Oliver Lindemann <oliver@expyriment.org>'
__version__ = ''
__revision__ = ''
__date__ = ''


def _empty_square(n):
    square = []
    for x in range(0, n):
        square.append([])
        for _i in range(0, n):
            square[x].append(None)
    return square


def _square_of_elements(list_, idx_square):
    """Return a square array of elements.

    Returns
    -------
    square : list
        a square array with the elements from the list defined by
        idx_square e.g.; idx_square[a][b] is the list index of element[a][b].

    """

    square = _empty_square(len(list_))
    for c in range(0, len(list_)):
        for r in range(0, len(list_)):
            square[r][c] = list_[idx_square[r][c]]
    return square


def is_permutation_type(type_str):
    """Return true if the string or value is a know permutation type.

    Parameters
    ----------
    type_str : string
        permutation type string

    """

    from ...misc import constants
    return (type_str == constants.P_RANDOM or \
            type_str == constants.P_CYCLED_LATIN_SQUARE or \
            type_str == constants.P_BALANCED_LATIN_SQUARE)


def _cycle_list(arr):
    rtn = arr[1:]
    rtn.append(arr[0])
    return rtn


def balanced_latin_square(elements):
    """A balanced latin square permutation of elements.

    If elements is an integer the elements=[0,..., elements] is used.

    Parameters
    ----------
    elements : int or list
        list of elements or a number

    """

    if isinstance(elements, (tuple, list)):
        idx = balanced_latin_square(len(elements))
        square = _square_of_elements(elements, idx)
    else:
        n = elements
        # Make n cycled columns [0,1,2,...n-1][1,2,3,...n-1,0][...]
        columns = cycled_latin_square(n)

        # Make index list to sort columns [0,1,n-1,3,n-2,4,...]
        c_idx = [0, 1]
        take_last = True
        tmp = list(range(2, n))
        for _i in range(2, n):
            if take_last:
                c_idx.append(tmp.pop())
            else:
                c_idx.append(tmp.pop(0))
            take_last = not take_last

        # Write sorted columns to square
        square = _empty_square(n)
        for c in range(0, n):
            for r in range(0, n):
                square[r][c] = columns[c_idx[c]][r]

    return square


def cycled_latin_square(elements):
    """A cycled latin square permutation of elements.

    If elements is a integer the elements=[0,..., elements] is used.

    Parameters
    ----------
    elements : int or list
        list of elements or a number

    """

    if isinstance(elements, (tuple, list)):
        idx = cycled_latin_square(len(elements))
        square = _square_of_elements(elements, idx)
    else:
        square = [list(range(0, elements))]
        for r in range(0, elements - 1):
            square.append(_cycle_list(square[r]))
    return square



#!/usr/bin/env python

"""
A Shape stimulus.

This module contains a class implementing a shape stimulus.

"""
from __future__ import absolute_import, print_function, division
from builtins import *


__author__ = 'Florian Krause <florian@expyriment.org>, \
Oliver Lindemann <oliver@expyriment.org>'
__version__ = ''
__revision__ = ''
__date__ = ''

import copy
from math import sqrt
import pygame

from . import defaults
from ._visual import Visual
from .. import _internals
from ..misc._timer import get_time
from ..misc.geometry import XYPoint, lines_intersect, position2coordinates


def _get_shape_rect(points):
    # helper function
    """ return bouncing rect as pygame.Rect rect (top, left, width, height) around the points."""

    t = 0
    l = 0
    r = 0
    b = 0

    for p in points:
        if p.x < l:
            l = p.x
        elif p.x > r:
            r = p.x
        if p.y > t:
            t = p.y
        elif p.y < b:
            b = p.y

    return pygame.Rect(l, t, r - l, t - b)

class Shape(Visual):
    """A class implementing a shape."""

    def __init__(self, position=None, colour=None,
                 line_width=None,
                 anti_aliasing=None,
                 vertex_list=None,
                 debug_contour_colour = None
                 ):

        """Create a shape.

        A shape is an object described by vertices. For more details about
        vertex representations see:
        http://en.wikipedia.org/wiki/Vertex_(geometry)

        The vertex representation describes the contour of an object. Think of
        it as if you would draw with a pen. You start somewhere and make
        movements in different directions. Shapes are always closed. Thus, a
        rectangle could be for instance described by a right, a down and a left.
        (The fours movement up to the origin is not required; see example below).

        A shapes is conceptually a filled object. The vertices describe
        the contour of the object. That is, a shape object has no
        line_width and has thus no contour of different colour. Displaying
        the contour might be useful for debugging purposes. To do so,
        set the parameter 'debug_contour_colour'. However, note that is
        affects the surface size inconsistently.

        If you want to plot a contour-like stimulus (e.g. a frame) the
        vertices need to describe the contour shape, in other words "the
        contour of to-be-displayed contour". Thus, a frame should be
        correct implemented as illustrated in example 2.

        As always in Expyriment, the center of the surface of the shape is its
        position. That means, that with every new vertex, the shape size might
        change and the shape position will be realigned.


        Examples
        --------
        EXAMPLE 1:
        >>> # drawing a rectangle with the size (100, 50)
        >>> r = stimuli.Shape()
        >>> r.add_vertices([(100,0), (0,-50),(-100,0) ])
        >>> # three vertices are sufficient, because shapes are always closed
        >>> r.present()

        EXAMPLE 2:
        >>> # drawing a frame
        >>> def vertices_frame(size, frame_thickness):
        >>>     # this function is also available in the geometry module
        >>>     return  [ (size[0]-frame_thickness, 0),
        >>>               (0, -size[1]),
        >>>               (-size[0], 0),
        >>>               (0, size[1]),
        >>>               (frame_thickness, 0),
        >>>               (0, -(size[1]-frame_thickness)),
        >>>               (size[0]-2*frame_thickness, 0),
        >>>               (0, size[1]-2*frame_thickness),
        >>>               (-(size[0]-2*frame_thickness), 0)]
        >>> fr = stimuli.Shape(vertex_list=vertices_frame(size=(200, 100),
        >>>               frame_thickness=10))
        >>> fr.present()

        EXAMPLE 3:
        >>> # using 'vertices_regular_polygon' from misc.geometry
        >>> from expyriment.misc import geometry
        >>> sh = stimuli.Shape(vertex_list=geometry.vertices_regular_polygon(5, 60))
        >>> sh.present()

        Parameters
        ----------
        position : (int, int), optional
            position of the stimulus
        colour : (int, int, int), optional
            colour of the shape or the fill colour
        anti_aliasing : int, optional
            anti aliasing parameter (good anti_aliasing with 10)
        vertex_list : (int, int)
            list of vertices (int, int)
        debug_contour_colour : (int, int, int), optional
            The colour of the contour of the shape.
            If None (default), contour colour is not displyed.
            Use this option only for debugging. The resulting surface size
            might be enlarged by one pixel large depending on the shape.

        Notes
        -----
        You can also use the convenience function `vertices_triangle`,
        `vertices_rectangle`, `vertices_regular_polygon`, `vertices_frame`
        defined in `expyriment.misc.geometry` to plot predefined geometrical
        shapes (see example 3).

        """

        if position is None:
            position = defaults.shape_position
        Visual.__init__(self, position)
        if colour is None:
            colour = defaults.shape_colour
        if colour is not None:
            self._colour = colour
        else:
            self._colour = _internals.active_exp.foreground_colour
        if anti_aliasing is not None:
            self._anti_aliasing = anti_aliasing
        else:
            self._anti_aliasing = defaults.shape_anti_aliasing

        self._debug_contour_colour = debug_contour_colour

        self._vertices = []
        self._xy_points = []
        self._rect = pygame.Rect(0, 0, 0, 0)
        self._native_rotation = 0
        self._native_scaling = [1, 1]
        self._native_rotation_centre = (0, 0)
        self._rotation_centre_display_colour = None
        self._update_points()

        if vertex_list is not None:
            self.add_vertices(vertex_list=vertex_list)

    _getter_exception_message = "Cannot perform {0} if surface exists!"

    def __repr__(self):
        return  "vertices: {0}; points: {1}".format(self.vertices,
                                                    self.points)

    @property
    def colour(self):
        """Getter for colour."""

        return self._colour

    @colour.setter
    def colour(self, colour):
        """Setter for colour."""

        if self.has_surface:
            raise AttributeError(Shape._getter_exception_message.format(
                "colour change"))
        self._colour = colour


    @property
    def anti_aliasing(self):
        """Getter for anti_aliasing."""

        return self._anti_aliasing

    @anti_aliasing.setter
    def anti_aliasing(self, value):
        """Setter for anti_aliasing."""

        if self.has_surface:
            raise AttributeError(Shape._getter_exception_message.format(
                "anti_aliasing"))
        self._anti_aliasing = value

    @property
    def rotation_centre(self):
        """Getter for rotation_centre."""
        return self._native_rotation_centre

    @rotation_centre.setter
    def rotation_centre(self, centre):
        """Setter for rotation_centre."""

        self._native_rotation_centre = centre
        self._update_points()


    @property
    def rotation_centre_display_colour(self):
        """Getter for rotation_centre_display_colour."""

        return self._rotation_centre_display_colour

    @rotation_centre_display_colour.setter
    def rotation_centre_display_colour(self, colour):
        """Setter for rotation_centre_display_colour.

        Set rotation_centre_display_colour to a colour (default=None) to
        display the centre of rotation.

        """

        self._rotation_centre_display_colour = colour
        self._update_points()

    @property
    def debug_contour_colour(self):
        """Getter for conture_colour."""

        return self._debug_contour_colour

    @debug_contour_colour.setter
    def debug_contour_colour(self, colour):
        """Setter for conture_colour."""

        if self.has_surface:
            raise AttributeError(Shape._getter_exception_message.format(
                "contour colour change"))
        self._debug_contour_colour = self.colour

    @property
    def width(self):
        return self._rect.width

    @property
    def height(self):
        return self._rect.height

    @property
    def shape_size(self):
        return self._rect.size

    @property
    def rect(self):
        """Getter for bouncing rectangular
        rect = pygame.Rect(left, top, width, height)

        Notes
        -----
        Fom version 0.9.1 on, this is a pygame.Rect

        """

        return self._rect

    @property
    def vertices(self):
        """Getter for the polygon vertices."""

        return self._vertices

    @property
    def points(self):
        """Returns shape as list of tuples (x, y) representing points.
         see Shape.xy_points

        The representation does not take into account the position. Use
        points_on_screen for position-depended representation.

        In contrast to the vertex representation, the point representation
        takes into all the native transformations (rotation, scaling,
        flipping)

        Returns
        -------
        val: list of tuples
            polygon as list of tuples (x,y) in Expyriment coordinates

        """

        rtn = []
        for p in self.xy_points:
            rtn.append(p.tuple)
        return rtn

    @property
    def points_on_screen(self):
        """Returns shape as list of tuples (x, y) resenting points on the
        screen. see Shape.xy_points_on_screen.

        In contrast to Shape.points, it takes into account the position on
        screen.

        In contrast to the vertex representation, the point representation
        takes into all the native transformations (rotation, scaling,
        flipping).

        Returns
        -------
        val: list of tuples
            polygon as list of tuples (x,y) in Expyriment coordinates

        """

        rtn = []
        for p in self.xy_points_on_screen:
            rtn.append(p.tuple)
        return rtn

    @property
    def scaling(self):
        """"Getter for the total native scaling."""
        return self._native_scaling

    @property
    def rotation(self):
        """"Getter for the total native rotation."""
        return self._native_rotation

    @property
    def flipping(self):
        """"Getter for the total native flipping."""
        return ((self.scaling[0] < 0), (self.scaling[1] < 0))

    @property
    def xy_points(self):
        """Returns the shape as list of XYPoints.

        The representation does not take into account the position. Use
        xy_points_on_screen for position-depended representation.

        In contrast to the vertex representation, the point representation
        takes into all the native transformations (rotation, scaling,
        flipping).

        Returns
        -------
        val: list of XYPoints
             polygon as list of XYPoints of the shape

        """

        return self._xy_points

    @property
    def xy_points_on_screen(self):
        """Returns the shape as list of XYPoints in Expyriment coordinates.

        In contrast to xy_points, it takes into account the position on
        screen.

        In contrast to the vertex representation, the point representation
        takes into all the native transformations (rotation, scaling,
        flipping).

        Returns
        -------
        val: list of XYPoints
             polygon as list of XYPoints of the shape

        """

        rtn = []
        pos = XYPoint(xy=self.position)
        for p in copy.deepcopy(self.xy_points):
            rtn.append(p.move(pos))
        return rtn

    def add_vertex(self, xy):
        """ Add a vertex to the shape.

        Parameters
        ----------
        xy : (int, int)
            vertex as tuple

        """

        self.add_vertices([xy])

    def add_vertices(self, vertex_list):
        """ Add a list of vertices to the shape.

        Parameters
        ----------
        vertex_list : ((int, int))
            list of vertices ((int, int))

        """
        type_error_message = "The method add_vertices requires a list of" + \
                            " tuples as argument."
        if self.has_surface:
            raise AttributeError(Shape._getter_exception_message.format(
                "add_vertices"))
        if not isinstance(vertex_list, (list, tuple)):
            raise TypeError(type_error_message)
        for xy in vertex_list:
            if not isinstance(xy, (list, tuple)):
                raise TypeError(type_error_message)
            if len(xy) != 2:
                raise TypeError(type_error_message)
            self._vertices.append(list(xy))
        self._update_points()

    def remove_vertex(self, index):
        """Remove a vertex."""

        if self.has_surface:
            raise AttributeError(Shape._getter_exception_message.format(
                "remove_vertex"))
        if index > 0 and index < len(self._vertices):
            self._vertices.pop(index)
        self._update_points()

    def erase_vertices(self):
        """Removes all vertices."""

        if self.has_surface:
            raise AttributeError(Shape._getter_exception_message.format(
                "erase_vertices"))

        self._vertices = []
        self._xy_points = []
        self._rect = pygame.Rect(0, 0, 0, 0)
        self._native_rotation = 0
        self._native_scaling = [1, 1]
        self._native_rotation_centre = (0, 0)
        self._rotation_centre_display_colour = None
        self._update_points()

    def convert_expyriment_xy_to_surface_xy(self, point_xy):
        """Convert a point from shape coordinates to surface coordinates.

        Parameters
        ----------
        point_xy : (int, int)
            Expyriment screen coordinates (tuple)

        """

        return (int(point_xy[0] - self._rect.left),
                - 1 * int(point_xy[1] - self._rect.top))

    def native_overlapping_with_position(self, position):
        """Return True if the position is inside the shape.

        Parameters
        position -- Expyriment screen coordinates (tuple)

        Returns
        -------
        val : bool
            True if the position is inside the shape

        """

        pt = XYPoint(position)
        return pt.is_inside_polygon(self.xy_points_on_screen)

    def is_point_inside(self, point_xy):
        """"OBSOLETE METHOD: Please use 'overlapping_with_position'."""

        raise DeprecationWarning("is_point_inside is an obsolete method. Please use overlapping_with_position")

    def overlapping_with_shape(self, other):
        """Return true if shape overlaps with other shape.

        Parameters
        ----------
        other : stimuli.Shape
            the other shape object

        Returns
        -------
        val : bool
            True if overlapping

        """

        # shape and other shape do not overlap if 
        # (a) no point of shape is inside other shape 
        # (b) AND no point of other shape is inside shape
        # (c) AND lines do not intersect

        s1 = self.xy_points_on_screen
        s2 = other.xy_points_on_screen
        #(a) & (b)
        for p in s1:
            if p.is_inside_polygon(s2):
                return True
        for p in s2:
            if p.is_inside_polygon(s1):
                return True
        #(c)
        for from1 in range(0, len(s1) - 1):
            for to1 in range(from1 + 1, len(s1)):
                for from2 in range(0, len(s2) - 1):
                    for to2 in range(0, len(s2)):
                        if lines_intersect(s1[from1], s1[to1], s2[from2], s2[to2]):
                            return True
        return False

    def is_shape_overlapping(self, shape2):
        """OBSOLETE METHOD: Please use 'overlapping_with_shape'."""

        raise DeprecationWarning("is_shape_overlapping is an obsolete method. Please use overlapping_with_shape.")

    def native_rotate(self, degree):
        """Rotate the shape.

        Native rotation of shape is a native operation (not a surface
        operation) and does therefore not go along with a quality loss.
        No surface will be created.

        Parameters
        ----------
        degree : float
            degree to rotate counterclockwise (float)

        """

        if self.has_surface:
            raise AttributeError(Shape._getter_exception_message.format(
                "native_rotate"))
        self._native_rotation = self._native_rotation + degree
        self._update_points()

    def native_scale(self, factors, scale_line_width=False):
        """Scale the shape.

        Native scaling of shapes is a native operation (not a surface
        operation) and does therefore not go along with a quality loss.
        No surface will be created.

        Negative scaling values will native_flip the stimulus.

        Parameters
        ----------
        factors : int or (int, int)
            x and y factors to scale

        """

        if not isinstance(factors, (list, tuple)):
            factors = [factors, factors]
        if self.has_surface:
            raise AttributeError(Shape._getter_exception_message.format(
                "native_scale"))

        self._native_scaling[0] = self._native_scaling[0] * factors[0]
        self._native_scaling[1] = self._native_scaling[1] * factors[1]
        self._update_points()

    def native_flip(self, booleans):
        """Flip the shape.

        Native flipping of shapes is a native operation (not a surface
        operation) and does therefore not go along with a quality loss.
        No surface will be created.

        Parameters
        ----------
        booleans : (bool, bool)
            booleans to flip horizontally and vertically or not

        """

        if self.has_surface:
            raise AttributeError(Shape._getter_exception_message.format(
                "native_flip"))
        if booleans[0]:
            self._native_scaling[0] = self._native_scaling[0] * -1
        if booleans[1]:
            self._native_scaling[1] = self._native_scaling[1] * -1
        self._update_points()

    def blur(self, level):
        """Blur the shape.

        This blurs the stimulus, by scaling it down and up by the factor of
        'level'.
        Notes
        -----
        Depending on the blur level and the size of your stimulus, this method
        may take some time!

        Parameters
        ----------
        level : int
            level of bluring

        Returns
        -------
        time : int
            the time it took to execute this method

        """

        start = get_time()
        self.scale((1.0 / level, 1.0 / level))
        self.scale((level, level))
        return int((get_time() - start) * 1000)


    def _update_points(self):
        """Updates the points of the shape and the drawing rect.

        Converts vertex to points, centers points, rotates, calculates rect
        """

        xy_p = [XYPoint(0, 0)]
        for v in self._vertices:
            # Copying and scaling and flipping of vertices
            v = (v[0] * self._native_scaling[0],
                 v[1] * self._native_scaling[1])
            # Converts tmp_vtx to points in xy-coordinates
            xy_p.append(XYPoint(x = v[0] + xy_p[-1].x,
                                y = v[1] + xy_p[-1].y))

        # center points
        r = _get_shape_rect(xy_p)
        cntr = (r.left + (r.width // 2), r.top - (r.height // 2))
        m = XYPoint(x=-1*cntr[0], y=-1*cntr[1])
        for x in range(0, len(xy_p)): # Center points
            xy_p[x].move(m)

        if self._native_rotation != 0:
            for x in range(0, len(xy_p)):
                xy_p[x].rotate(self._native_rotation,
                               self._native_rotation_centre)

        self._xy_points = xy_p
        self._rect = _get_shape_rect(xy_p)

    def _create_surface(self):
        """Create the surface of the stimulus."""

        # Trick: draw enlarged shape and reduce size
        if self._anti_aliasing > 0: # Draw enlarged shape
            aa_scaling = (self._anti_aliasing / 5.0) + 1
            old_scaling = copy.copy(self._native_scaling)
            self.native_scale([aa_scaling, aa_scaling])

        if self.debug_contour_colour is None:
            line_width = 0 # change if contours
        else:
            line_width = 1 # change if contours

        # make surface
        target_surface_size = (self.width  + line_width, self.height + line_width)
        surface = pygame.surface.Surface(target_surface_size,
                                        pygame.SRCALPHA).convert_alpha()

        # plot create polygon area
        poly = []
        for p in self.xy_points: # Convert points_in_pygame_coordinates
            poly.append(self.convert_expyriment_xy_to_surface_xy(p.tuple))

        pygame.draw.polygon(surface, self.colour, poly, 0)

        # plot polygon contours manually
        if self.debug_contour_colour is not None:
            pygame.draw.polygon(surface, self.debug_contour_colour, poly, 1)

        if self._rotation_centre_display_colour is not None:
            rot_centre = self.convert_expyriment_xy_to_surface_xy(
                self._native_rotation_centre)
            pygame.draw.circle(surface, self._rotation_centre_display_colour,
                               rot_centre, 2)

        if self._anti_aliasing > 0: # Scale back
            size = surface.get_size()
            surface = pygame.transform.smoothscale(surface,
                                (int(size[0] / aa_scaling),
                                 int(size[1] / aa_scaling)))
            self._native_scaling = old_scaling
            self._update_points()

        return surface

    @staticmethod
    def _demo(exp=None):
        if exp is None:
            from .. import control
            control.set_develop_mode(True)
            control.defaults.event_logging = 0
            exp_ = control.initialize()
        sh = Shape(position=(10, 100), colour=(255, 0, 255))
        sh.add_vertices([(60, 60), (0, -120)])
        sh.present()
        if exp is None:
            exp_.clock.wait(1000)



"""
The statistics module.

This module contains miscellaneous stastistical functions for expyriment.

"""

__author__ = 'Florian Krause <florian@expyriment.org>, \
Oliver Lindemann <oliver@expyriment.org>'
__version__ = ''
__revision__ = ''
__date__ = ''


from ._statistics import sum, mode, mean, median, variance, std
from ._statistics import frequence_table, frequency_table



"""The design package.

This package provides several data structures for describing the design of an
experiment.  See also expyriment.design.extras for more design.

"""

__author__ = 'Florian Krause <florian@expyriment.org> \
              Oliver Lindemann <oliver@expyriment.org>'
__version__ = ''
__revision__ = ''
__date__ = ''


from . import defaults
from . import permute
from . import randomize
from ._structure import Experiment, Block, Trial

from .. import _internals
_internals.active_exp = Experiment("None")

